# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides class for analyzing complex network embedded on a spherical surface.
"""

from typing import Tuple
from collections.abc import Hashable

import numpy as np
from numpy import random
import igraph

from .spatial_network import SpatialNetwork
from .geo_grid import GeoGrid


class GeoNetwork(SpatialNetwork):
    """
    Encapsulates a network embedded on a spherical surface.

    Particularly adds more network measures and statistics based on the
    spatial embedding.

    :ivar node_weight_type: (string) - The type of geographical node weight to
                            be used.
    """

    #
    #  Definitions of internal methods
    #

    def __init__(self, grid: GeoGrid, adjacency=None, edge_list=None,
                 directed=False, node_weight_type="surface", silence_level=0):
        """
        Initialize an instance of GeoNetwork.

        :type grid: :class:`.GeoGrid`
        :arg grid: The GeoGrid object describing the network's spatial
            embedding.
        :type adjacency: 2D array (int8) [index, index]
        :arg adjacency: The network's adjacency matrix.
        :type edge_list: array-like list of lists
        :arg  edge_list: Edge list of the new network.
                         Entries [i,0], [i,1] contain the end-nodes of an edge.
        :arg bool directed: Determines, whether the network is treated as
            directed.
        :arg str node_weight_type: The type of geographical node weight to be
            used.
        :arg int silence_level: The inverse level of verbosity of the object.

        Possible choices for ``node_weight_type``:
          - None (constant unit weights)
          - "surface" (cos lat)
          - "irrigation" (cos² lat)
        """
        assert isinstance(grid, GeoGrid)
        self.grid: GeoGrid = grid
        """GeoGrid object describing the network's spatial embedding"""

        #  Call constructor of parent class Network
        SpatialNetwork.__init__(self, grid=grid, adjacency=adjacency,
                                edge_list=edge_list, directed=directed,
                                silence_level=silence_level)

        #  Set area weights
        self.set_node_weight_type(node_weight_type)

        #  cartesian coordinates of nodes
        self.cartesian = None
        self.grid_neighbours = None
        self.grid_neighbours_set = None

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        # The following attributes are assumed immutable:
        #   (grid)
        return SpatialNetwork.__cache_state__(self)

    def __rec_cache_state__(self) -> Tuple[object, ...]:
        return (self.grid,)

    def __str__(self):
        """
        Return a string representation of the GeoNetwork object.
        """
        return (f'GeoNetwork:\n{SpatialNetwork.__str__(self)}\n'
                f'Geographical boundaries:\n{self.grid.print_boundaries()}')

    def set_node_weight_type(self, node_weight_type):
        """
        Set node weights for calculation of n.s.i. measures according to
        requested type.

        Possible choices for ``node_weight_type``:
          - None (constant unit weights)
          - "surface" (cos lat)
          - "irrigation" (cos² lat)

        :arg str node_weight_type: The type of geographical node weight to be
            used.
        """
        if self.silence_level <= 1:
            print("Setting area weights according to type "
                  f"{node_weight_type} ...")

        #  Set instance variable accordingly
        self.node_weight_type = node_weight_type

        if node_weight_type == "surface":
            self._node_weights = self.grid.cos_lat()
        elif node_weight_type == "irrigation":
            self._node_weights = np.square(self.grid.cos_lat())
        #  If None or invalid choice:
        else:
            self._node_weights = None

    #
    #  Load and save GeoNetwork object
    #

    @staticmethod
    def Load(filename, fileformat=None, silence_level=0, *args, **kwds):
        """
        Return a GeoNetwork object stored in files.

        Unified reading function for graphs. Relies on and partially extends
        the corresponding igraph function. Refer to igraph documentation for
        further details on the various reader methods for different formats.

        This method tries to identify the format of the graph given in
        the first parameter and calls the corresponding reader method.

        Existing node and link attributes/weights are also restored depending
        on the chosen file format. E.g., the formats GraphML and gzipped
        GraphML are able to store both node and link weights.

        The remaining arguments are passed to the reader method without
        any changes.

        :arg tuple/list filename: Tuple or list of two strings, namely
            the paths to the files containing the Network object
            and the GeoGrid object (filename_network, filename_grid)
        :arg str fileformat: the format of the file (if known in advance)
          ``None`` means auto-detection. Possible values are: ``"ncol"`` (NCOL
          format), ``"lgl"`` (LGL format), ``"graphml"``, ``"graphmlz"``
          (GraphML and gzipped GraphML format), ``"gml"`` (GML format),
          ``"net"``, ``"pajek"`` (Pajek format), ``"dimacs"`` (DIMACS format),
          ``"edgelist"``, ``"edges"`` or ``"edge"`` (edge list),
          ``"adjacency"`` (adjacency matrix), ``"pickle"`` (Python pickled
          format).
        :arg int silence_level: The inverse level of verbosity of the object.
        :rtype: SpatialNetwork object
        :return: :class:`GeolNetwork` instance.
        """
        try:
            (filename_network, filename_grid) = filename
        except ValueError as e:
            raise ValueError("'filename' must be a tuple or list of two "
                             "items: filename_network, filename_grid") from e

        #  Load Grid object
        grid = GeoGrid.Load(filename_grid)
        print(grid.__class__)

        #  Load to igraph Graph object
        graph = igraph.Graph.Read(f=filename_network, format=fileformat,
                                  *args, **kwds)

        #  Extract adjacency matrix
        A = np.array(graph.get_adjacency(type=2).data)

        #  Create GeoNetwork instance
        net = GeoNetwork(adjacency=A, grid=grid,
                         directed=graph.is_directed(),
                         silence_level=silence_level)

        #  Extract node weights
        if "node_weight_nsi" in graph.vs.attribute_names():
            node_weights = \
                np.array(graph.vs.get_attribute_values("node_weight_nsi"))
            net.node_weights = node_weights

        #  Overwrite igraph Graph object in Network instance to restore link
        #  attributes/weights
        net.graph = graph
        #  invalidate cache
        net._mut_la += 1
        return net

    def save_for_cgv(self, filename, fileformat="graphml"):
        """
        Save the GeoNetwork and its attributes for the CGV visualization
        software.

        The node coordinates are stored as node attributes by default, likewise
        angular link distances are stored as edge attributes by default. All
        additional node and link properties are also stored for visualization.

        This format is intended for being used by the spatial graph
        visualization software CGV developed in Rostock (contact Thomas Nocke,
        nocke@pik-potsdam.de). By default, the file includes the latitude and
        longitude vectors as node properties, as well as the geodesic angular
        distance as an link property.

        :arg str file_name: The file name should end with ".dot" or ".gml".
        :arg str fileformat: The file format: "graphml"  - GraphML format
            "graphmlz" - gzipped GraphML format
            "graphviz" - GraphViz format
        """
        #  Save node coordinates as node attribute
        self.set_node_attribute("lat", self.grid.lat_sequence())
        self.set_node_attribute("lon", self.grid.lon_sequence())

        #  Save geodesic angular distances on the sphere as link attribute
        self.set_link_attribute("ang_dist", self.grid.angular_distance())

        #  Save network, independent of filename!
        if fileformat in ["graphml", "graphmlz", "graphviz"]:
            self.graph.save(filename, format=fileformat)
        else:
            print("ERROR: the chosen format is not supported by save_for_cgv "
                  "for use with the CGV software.")

    #
    #  Graph generation methods
    #

    @staticmethod
    def SmallTestNetwork():
        """
        Return a 6-node undirected geographically embedded test network.

        The test network consists of the SmallTestNetwork of the Network class
        with node coordinates given by the SmallTestGrid of the GeoGrid class.

        The network looks like this::

                3 - 1
                |   | \\
            5 - 0 - 4 - 2

        :rtype: GeoNetwork instance
        :return: an instance of GeoNetwork for testing purposes.
        """
        return GeoNetwork(grid=GeoGrid.SmallTestGrid(),
                          adjacency=SpatialNetwork.SmallTestNetwork()
                          .adjacency,
                          directed=False, node_weight_type="surface",
                          silence_level=2)

    @staticmethod
    def Model(network_model, grid, node_weight_type="surface", **kwargs):
        """
        Return a new model graph generated with the specified network model
        and embedded on a geographical grid
        """
        A = getattr(GeoNetwork, network_model)(**kwargs)
        return GeoNetwork(adjacency=A, grid=grid, directed=False,
                          node_weight_type=node_weight_type)

    #
    #  Graph randomization methods
    #

    #  FIXME: Check this method and implement in C++ via Cython for speed.
    #  FIXME: Also improve documentation.
    #  FIXME: Add example
    def shuffled_by_distance_copy(self):
        """
        Return a copy of the network where all links in each node-distance
        class have been randomly re-assigned.

        In other words, the result is a random network in which the link
        probability only depends on the nodes' distance and is the same as in
        the original network.

        :rtype: GeoNetwork
        :return: the distance shuffled copy.
        """
        N = self.N
        A = self.adjacency
        D = self.grid.distance()

        #  Count pairs and links by distance
        n_pairs_by_dist = {}
        n_links_by_dist = {}
        for j in range(0, N):
            print(j)
            for i in range(0, j):
                d = D[i, j]
                try:
                    n_pairs_by_dist[d] += 1
                except KeyError:
                    n_pairs_by_dist[d] = 1
                if A[i, j]:
                    try:
                        n_links_by_dist[d] += 1
                    except KeyError:
                        n_links_by_dist[d] = 1

        #  Determine link probabilities
        p_by_dist = {}
        for d, n in n_pairs_by_dist.items():
            try:
                p_by_dist[d] = n_links_by_dist[d] * 1.0 / n
            except KeyError:
                p_by_dist[d] = 0.0
            print(d, p_by_dist[d])
        del n_links_by_dist, n_pairs_by_dist

        #  Link new pairs with respective probability
        A_new = np.zeros((N, N))
        for j in range(0, N):
            print("new ", j)
            for i in range(0, j):
                d = D[i, j]
                if p_by_dist[d] >= np.random.random():
                    A_new[i, j] = A_new[j, i] = 1
                    print(i, j, d, p_by_dist[d])

        #  Create new GeoNetwork object based on A_new
        net = GeoNetwork(adjacency=A_new, grid=self.grid,
                         directed=self.directed,
                         node_weight_type=self.node_weight_type,
                         silence_level=self.silence_level)

        return net

    #
    #  Generate a geographical distribution
    #

    #  FIXME: Derive this method from a generalized variant based on n.s.i.
    #  distributions.
    def geographical_distribution(self, sequence, n_bins):
        """
        Return a normalized geographical frequency distribution.

        Also return the estimated statistical error and lower bin boundaries.

        This function counts which percentage of total surface area falls into
        a bin and NOT which number of nodes does so.

        .. note::
           Be aware that this method only returns meaningful results
           for regular rectangular grids, where the representative area of each
           node is proportional to the cosine of its latitude.

        **Example:**

        >>> net = GeoNetwork.SmallTestNetwork()
        >>> r(net.geographical_distribution(
        ...     sequence=net.degree(), n_bins=3)[0])
        array([ 0.1565, 0.3367, 0.5068])

        :type sequence: 1D Numpy array [index]
        :arg sequence: The input sequence (e.g., some local network measure).

        :type n_bins: number (int)
        :arg n_bins: The number of bins for histogram.

        :rtype: tuple of three 1D Numpy arrays [bin]
        :return: the geographical distribution, statistical error, and lower
                 bin boundaries.
        """
        if self.silence_level <= 1:
            print("Calculating geographical frequency distribution...")

        #  Initializations
        hist = np.zeros(n_bins)

        #  Get sequence of cosines of latitude
        cos_lat = self.grid.cos_lat()
        #  Calculate total dimensionless area of the sphere
        norm = cos_lat.sum()

        #  Get range for histogram
        range_min = float(sequence.min())
        range_max = float(sequence.max())

        #  Calculate scaling factor for histogram
        scaling = 1. / (range_max - range_min)

        #  Get array of symbols corresponding to sequence
        symbolic = \
            ((n_bins - 1) * scaling * (sequence - range_min)).astype("int")

        #  Calculate histogram
        for i in range(len(sequence)):
            hist[symbolic[i]] += cos_lat[i]

        #  Normalize histogram by the total dimensionless area
        hist /= norm

        #  Construct lower bin boundaries
        lbb = np.linspace(range_min, range_max, n_bins + 1)[:-1]

        #  Calculate statistical error given by 1/n_i per bin i,
        #  where n_i is the number of samples per bin.
        error = np.zeros(n_bins)
        error[hist != 0] = 1 / np.sqrt(hist[hist != 0])

        return (hist, error, lbb)

    def geographical_cumulative_distribution(self, sequence, n_bins):
        """
        Return a normalized geographical cumulative distribution.

        Also return estimated statistical error and the lower bin boundaries.

        This function counts which percentage of total surface area has a value
        of sequence larger or equal than the one bounded by a specific bin and
        NOT which number of nodes does so.

        .. note::
           Be aware that this method only returns meaningful results
           for regular rectangular grids, where the representative area of each
           node is proportional to the cosine of its latitude.

        **Example:**

        >>> net = GeoNetwork.SmallTestNetwork()
        >>> r(net.geographical_cumulative_distribution(
        ...     sequence=net.degree(), n_bins=3)[0])
        array([ 1. , 0.8435, 0.5068])

        :type sequence: 1D Numpy array [index]
        :arg sequence: The input sequence (e.g., some local network measure).

        :type n_bins: number (int)
        :arg n_bins: The number of bins for histogram.

        :rtype: tuple of three 1D Numpy arrays [bin]
        :return: the cumulative geographical distribution, statistical error,
                 and lower bin boundaries.
        """
        (dist, error, lbb) = self.geographical_distribution(sequence, n_bins)
        cumu_dist = np.zeros(n_bins)
        for i in range(n_bins):
            cumu_dist[i] = dist[i:].sum()
        return (cumu_dist, error, lbb)

    #
    #  Area weighted connectivity (AWC) related measures
    #

    def area_weighted_connectivity(self):
        """
        Return area weighted connectivity (:math:`AWC`).

        It gives the fractional area of the network, a node is connected to.
        :math:`AWC` is closely related to node splitting invariant degree
        :meth:`.Network.nsi_degree` with area as node weight.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().area_weighted_connectivity())
        array([ 0.4854, 0.499 , 0.3342, 0.3446, 0.5146, 0.1726])

        :rtype: 1D Numpy array [index]
        :return: the area weighted connectivity sequence.
        """
        if self.silence_level <= 1:
            print("Calculating area weighted connectivity...")

        if self.directed:
            return (self.inarea_weighted_connectivity()
                    + self.outarea_weighted_connectivity())
        else:
            return self.inarea_weighted_connectivity()

    def inarea_weighted_connectivity(self):
        """
        Return in-area weighted connectivity.

        It gives the fractional area of the netwerk that connects to a given
        node. For undirected networks, it calculates total area weighted
        connectivity.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().inarea_weighted_connectivity())
        array([ 0.4854, 0.499 , 0.3342, 0.3446, 0.5146, 0.1726])

        :rtype: 1D Numpy array [index]
        :return: the in-area weighted connectivity sequence.
        """
        if self.silence_level <= 1:
            print("Calculating in - area weighted connectivity...")

        #  Calculate the sequence of cosine of latitude for all nodes
        cos_lat = self.grid.cos_lat()

        #  Calculate total dimensionless area of the sphere
        norm = cos_lat.sum()

        #  Normalize area weighted connectivity by the total dimensionless area
        inawc = cos_lat.dot(self.adjacency) / norm

        return inawc

    def outarea_weighted_connectivity(self):
        """
        Return out-area weighted connectivity.

        It gives the fractional area of the netwerk that a given node connects
        to. For undirected networks, it calculates total area weighted
        connectivity.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                outarea_weighted_connectivity())
        array([ 0.4854, 0.499 , 0.3342, 0.3446, 0.5146, 0.1726])

        :rtype: 1D Numpy array [index]
        :return: the out-area weighted connectivity sequence.
        """
        if self.silence_level <= 1:
            print("Calculating out - area weighted connectivity...")

        #  Calculate the sequence of cosine of latitude for all nodes
        cos_lat = self.grid.cos_lat()

        #  Calculate total dimensionless area of the sphere
        norm = cos_lat.sum()

        #  Normalize area weighted connectivity by the total dimensionless area
        outawc = np.dot(self.adjacency, cos_lat) / norm

        return outawc

    def area_weighted_connectivity_distribution(self, n_bins):
        """
        Return the area weighted connectivity frequency distribution.

        Also return estimated statistical error and lower bin boundaries.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                area_weighted_connectivity_distribution(n_bins=4)[0])
        array([ 0.1565, 0.3367, 0.3446, 0.1622])

        :type n_bins: number (int)
        :arg n_bins: The number of bins for histogram.

        :rtype: tuple of three 1D Numpy arrays [bin]
        :return: the :math:`AWC` distribution, statistical error,
                 and lower bin boundaries.
        """
        if self.silence_level <= 1:
            print("Calculating AWC frequency distribution...")

        awc = self.area_weighted_connectivity()

        return self.geographical_distribution(awc, n_bins)

    def inarea_weighted_connectivity_distribution(self, n_bins):
        """
        Return the in-area weighted connectivity frequency distribution.

        Also return estimated statistical error and lower bin boundaries.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                inarea_weighted_connectivity_distribution(n_bins=4)[0])
        array([ 0.1565, 0.3367, 0.3446, 0.1622])

        :type n_bins: number (int)
        :arg n_bins: The number of bins for histogram.

        :rtype: tuple of three 1D Numpy arrays [bin]
        :return: the in-:math:`AWC` distribution, statistical error,
                 and lower bin boundaries.
        """
        if self.silence_level <= 1:
            print("Calculating in-AWC frequency distribution...")

        in_awc = self.inarea_weighted_connectivity()

        return self.geographical_distribution(in_awc, n_bins)

    def outarea_weighted_connectivity_distribution(self, n_bins):
        """
        Return the out-area weighted connectivity frequency distribution.

        Also return estimated statistical error and lower bin boundaries.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                outarea_weighted_connectivity_distribution(n_bins=4)[0])
        array([ 0.1565, 0.3367, 0.3446, 0.1622])

        :type n_bins: number (int)
        :arg n_bins: The number of bins for histogram.

        :rtype: tuple of three 1D Numpy arrays [bin]
        :return: the out-:math:`AWC` distribution, statistical error,
                 and lower bin boundaries.
        """
        if self.silence_level <= 1:
            print("Calculating out-AWC frequency distribution...")

        out_awc = self.outarea_weighted_connectivity()

        return self.geographical_distribution(out_awc, n_bins)

    def area_weighted_connectivity_cumulative_distribution(self, n_bins):
        """
        Return the cumulative area weighted connectivity distribution.

        Also return estimated statistical error and lower bin boundaries.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                area_weighted_connectivity_cumulative_distribution(
        ...         n_bins=4)[0])
        array([ 1. , 0.8435, 0.5068, 0.1622])

        :type n_bins: number (int)
        :arg n_bins: The number of bins for histogram.

        :rtype: tuple of three 1D Numpy arrays [bin]
        :return: the cumulative :math:`AWC` distribution, statistical error,
                 and lower bin boundaries.
        """
        if self.silence_level <= 1:
            print("Calculating cumulative AWC distribution...")

        awc = self.area_weighted_connectivity()

        return self.geographical_cumulative_distribution(awc, n_bins)

    def inarea_weighted_connectivity_cumulative_distribution(self, n_bins):
        """
        Return the cumulative in-area weighted connectivity distribution.

        Also return estimated statistical error and lower bin boundaries.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                inarea_weighted_connectivity_cumulative_distribution(
        ...         n_bins=4)[0])
        array([ 1. , 0.8435, 0.5068, 0.1622])

        :type n_bins: number (int)
        :arg n_bins: The number of bins for histogram.

        :rtype: tuple of three 1D Numpy arrays [bin]
        :return: the cumulative in-:math:`AWC` distribution, statistical error,
                 and lower bin boundaries.
        """
        if self.silence_level <= 1:
            print("Calculating cumulative in-AWC distribution...")

        in_awc = self.inarea_weighted_connectivity()

        return self.geographical_cumulative_distribution(in_awc, n_bins)

    def outarea_weighted_connectivity_cumulative_distribution(self, n_bins):
        """
        Return the cumulative out-area weighted connectivity distribution.

        Also return estimated statistical error and lower bin boundaries.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                outarea_weighted_connectivity_cumulative_distribution(
        ...         n_bins=4)[0])
        array([ 1. , 0.8435, 0.5068, 0.1622])

        :type n_bins: number (int)
        :arg n_bins: The number of bins for histogram.

        :rtype: tuple of three 1D Numpy arrays [bin]
        :return: the cumulative out-:math:`AWC` distribution, statistical
                 error, and lower bin boundaries.
        """
        if self.silence_level <= 1:
            print("Calculating cumulative out-AWC distribution...")

        out_awc = self.outarea_weighted_connectivity()

        return self.geographical_cumulative_distribution(out_awc, n_bins)

    def average_neighbor_area_weighted_connectivity(self):
        """
        Return average neighbor area weighted connectivity.

        .. note::
           Does not use directionality information.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                average_neighbor_area_weighted_connectivity())
        array([ 0.3439, 0.3978, 0.5068, 0.4922, 0.4395, 0.4854])

        :rtype: 1D Numpy array [index]
        :return: the average neighbor area weighted connectivity sequence.
        """
        if self.silence_level <= 1:
            print("Calculating average neighbour AWC...")

        A = self.undirected_adjacency()
        degree = self.degree()
        awc = self.area_weighted_connectivity()
        average_neighbor_awc = A * awc

        #  Normalize by node degree
        average_neighbor_awc[degree != 0] /= degree[degree != 0]
        return average_neighbor_awc

    def max_neighbor_area_weighted_connectivity(self):
        """
        Return maximum neighbor area weighted connectivity.

        .. note::
           Does not use directionality information.

        >>> r(GeoNetwork.SmallTestNetwork().\
                max_neighbor_area_weighted_connectivity())
        array([ 0.5146, 0.5146, 0.5146, 0.499 , 0.499 , 0.4854])

        :rtype: 1D Numpy array [index]
        :return: the maximum neighbor area weighted connectivity sequence.
        """
        if self.silence_level <= 1:
            print("Calculating maximum neighbour AWC...")

        A = self.undirected_adjacency().A
        awc = self.area_weighted_connectivity()
        max_neighbor_awc = np.zeros(self.N)

        for i in range(self.N):
            max_neighbor_awc[i] = awc[A[i, :] == 1].max()

        return max_neighbor_awc

    #
    #  Distance related measures
    #

    #  (Un)directed total link distances

    def total_link_distance(self, geometry_corrected=False):
        """
        Return the sequence of total link distances for all nodes.

        .. note::
           Does not use directionality information.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                total_link_distance(geometry_corrected=False))
        array([ 0.1886, 0.097 , 0.0486, 0.0838, 0.1498, 0.0837])

        :arg bool geometry_corrected: Toggles geometry correction.
        :rtype: 1D array [index]
        :return: the total link distance sequence.
        """
        if self.silence_level <= 1:
            print("Calculating total link distance...")

        ald = self.average_link_distance(geometry_corrected)
        awc = self.area_weighted_connectivity()

        return ald * awc

    def intotal_link_distance(self, geometry_corrected=False):
        """
        Return the sequence of in-total link distances for all nodes.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                intotal_link_distance(geometry_corrected=False))
        array([ 0.1886, 0.097 , 0.0486, 0.0838, 0.1498, 0.0837])

        :arg bool geometry_corrected: Toggles geometry correction.
        :rtype: 1D array [index]
        :return: the in-total link distance sequence.
        """
        if self.silence_level <= 1:
            print("Calculating in-total link distance...")

        in_ald = self.inaverage_link_distance(geometry_corrected)
        in_awc = self.inarea_weighted_connectivity()

        return in_ald * in_awc

    def outtotal_link_distance(self, geometry_corrected=False):
        """
        Return the sequence of out-total link distances for all nodes.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                outtotal_link_distance(geometry_corrected=False))
        array([ 0.1886, 0.097 , 0.0486, 0.0838, 0.1498, 0.0837])

        :arg bool geometry_corrected: Toggles geometry correction.
        :rtype: 1D array [index]
        :return: the out-total link distance sequence.
        """
        if self.silence_level <= 1:
            print("Calculating out-total link distance...")

        out_ald = self.outaverage_link_distance(geometry_corrected)
        out_awc = self.outarea_weighted_connectivity()

        return out_ald * out_awc

    #  (Un)directed connectivity weighted link distances

    def _calculate_general_connectivity_weighted_distance(self, adjacency,
                                                          degrees):
        """
        Return general connectivity weighted link distances (CWD).

        This method is called to calculate undirected CWD, in-CWD
        and out-CWD.

        :type adjacency: 2D array [index, index]
        :arg adjacency: The adjacency matrix.
        :type degrees: 1D array [index]
        :arg degrees: The degree sequence.
        :rtype: 1D array [index]
        :return: the general connectivity weighted distance sequence.
        """
        D = self.grid.angular_distance()
        connectivity_weighted_distance = np.zeros(self.N)

        cos_lat = self.grid.cos_lat()
        norm = cos_lat.sum()

        for i in range(self.N):
            connectivity_weighted_distance[i] = \
                (adjacency[i, :] * cos_lat * D[i, :]).sum()

        #  Normalize by node degree and total dimensionless area
        connectivity_weighted_distance[degrees != 0] /= \
            degrees[degrees != 0] * norm

        return connectivity_weighted_distance

    def connectivity_weighted_distance(self):
        """
        Return undirected connectivity weighted link distances (CWD).

        .. note::
           Does not use directionality information.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                connectivity_weighted_distance())
        array([ 0.0625, 0.0321, 0.0241, 0.0419, 0.05 , 0.0837])

        :rtype: 1D Numpy array [index]
        :return: the undirected connectivity weighted distance sequence.
        """
        if self.silence_level <= 1:
            print("Calculating connectivity weighted link distance...")

        A = self.undirected_adjacency().A
        degree = self.degree()
        return self._calculate_general_connectivity_weighted_distance(
            A, degree)

    def inconnectivity_weighted_distance(self):
        """
        Return in-connectivity weighted link distances (CWD).

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                inconnectivity_weighted_distance())
        array([ 0.0625, 0.0321, 0.0241, 0.0419, 0.05 , 0.0837])

        :rtype: 1D Numpy array [index]
        :return: the in-connectivity weighted distance sequence.
        """
        if self.silence_level <= 1:
            print("Calculating in-connectivity weighted link distance...")

        A = self.adjacency.transpose()
        indegree = self.indegree()

        return self._calculate_general_connectivity_weighted_distance(A,
                                                                      indegree)

    def outconnectivity_weighted_distance(self):
        """
        Return out-connectivity weighted link distances (CWD).

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                outconnectivity_weighted_distance())
        array([ 0.0625, 0.0321, 0.0241, 0.0419, 0.05 , 0.0837])

        :rtype: 1D Numpy array [index]
        :return: the out-connectivity weighted distance sequence.
        """
        if self.silence_level <= 1:
            print("Calculating out-connectivity weighted link distance...")

        A = self.adjacency
        outdegree = self.outdegree()

        return self._calculate_general_connectivity_weighted_distance(
            A, outdegree)

    #
    #  Clustering coefficients including geographical information
    #

    def local_geographical_clustering(self):
        """
        Return local geographical clustering.

        Returns the sequence of local clustering coefficients weighted by the
        inverse angular great circle distance between nodes. This guarantees,
        that short links between spatially neighboring nodes in a triangle are
        weighted higher than long links between nodes that are spatially
        far away.

        Uses a definition of weighted clustering coefficient introduced in
        [Holme2007]_.

        .. note::
           Experimental measure!

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().local_geographical_clustering())
        Calculating local weighted clustering coefficient...
        array([ 0. , 0.0998, 0.1489, 0. , 0.2842, 0. ])

        :rtype: 1D Numpy array (index)
        :return: the local geographical clustering sequence.
        """
        ad = self.grid.angular_distance().copy()
        ad[ad == 0] = np.inf
        distance_weighted_adjacency = self.adjacency / ad
        return self.weighted_local_clustering(distance_weighted_adjacency)

    #  TODO: Experimental code!
    #  TODO: Improve documentation (Jobst).
    def nsi_connected_hamming_cluster_tree(self, lon_closed=True,
                                           lat_closed=False, alpha=0.01):
        """
        Perform NSI agglomerative clustering.

        Minimize in each step the Hamming distance between the original and
        the clustered network, but only joins connected clusters.

        Return c,h where c[i,j] = i iff node j is in cluster no. i,
        and 0 otherwise, and h is the corresponding list of total resulting
        relative Hamming distance between 0 and 1. The cluster numbers for all
        nodes and a k clusters solution is then c[:2*N-k,:].max(axis=0)

        :arg bool lon_closed: TODO
        :arg bool lat_closed: TODO
        :arg float alpha: TODO

        :rtype: TODO
        :return: TODO
        """
        N = self.N
        B = np.zeros((N, N)).astype("int")
        width = self.grid.grid()["lon"].size

        for i in range(0, N):
            if i % width > 0:
                B[i, i - 1] = 1
            elif lon_closed:
                B[i, i - 1 + width] = 1
            if i % width < width - 1:
                B[i, i + 1] = 1
            elif lon_closed:
                B[i, i + 1 - width] = 1
            if i >= width:
                B[i, i - width] = 1
            elif lat_closed:
                B[i, i - width + N] = 1
            if i < N - width:
                B[i, i + width] = 1
            elif lat_closed:
                B[i, i + width - N] = 1

        return self.do_nsi_hamming_clustering(admissible_joins=B, alpha=alpha)

    @staticmethod
    def cartesian2latlon(pos):
        return np.arcsin(pos[2]) * 180 / np.pi, \
            np.arctan2(pos[0], pos[1]) * 180 / np.pi

    @staticmethod
    def latlon2cartesian(lat, lon):
        lat *= np.pi/180
        lon *= np.pi/180
        coslat = np.cos(lat)
        return [coslat * np.sin(lon), coslat * np.cos(lon), np.sin(lat)]

    def boundary(self, nodes, geodesic=True, gap=0.0):
        """
        Return a list of ordered lists of nodes on the connected parts of the
        boundary of a subset of nodes and a list of ordered lists of (lat,lon)
        coordinates of the corresponding polygons

        * EXPERIMENTAL! *
        """
        #  Optional import for this experimental method
        try:
            import stripack  # @UnresolvedImport
            # tries to import stripack.so which must have been compiled with
            # f2py -c -m stripack stripack.f90
        except ImportError as err:
            raise RuntimeError("NOTE: stripack.so not available, "
                               "boundary() won't work.") from err

        N = self.N
        nodes_set = set(nodes)
        if len(nodes_set) >= N:
            return [], [], [], [(0.0, 0.0)]
        # find grid neighbours:
        if geodesic:
            if self.cartesian is not None:
                pos = self.cartesian
            else:
                # find cartesian coordinates of nodes,
                # assuming a perfect unit radius sphere:
                lat = self.grid.lat_sequence() * np.pi / 180
                lon = self.grid.lon_sequence() * np.pi / 180
                pos = self.cartesian = np.zeros((N, 3))
                coslat = np.cos(lat)
                self.cartesian[:, 0] = coslat * np.sin(lon)
                self.cartesian[:, 1] = coslat * np.cos(lon)
                self.cartesian[:, 2] = np.sin(lat)

                # find neighbours of each node in Delaunay triangulation,
                # sorted in counter-clockwise order, using stripack fortran
                # library:
                #  will contain 1-based node indices
                list_ = np.zeros(6*(N-2)).astype("int32")
                #  will contain 1-based list_ indices
                lptr = np.zeros(6*(N-2)).astype("int32")
                #  will contain 1-based list_ indices
                lend = np.zeros(N).astype("int32")
                lnew = 0
                near = np.zeros(N).astype("int32")
                foll = np.zeros(N).astype("int32")
                dist = np.zeros(N)
                ier = 0
                stripack.trmesh(self.cartesian[:, 0],
                                self.cartesian[:, 1],
                                self.cartesian[:, 2],
                                list_, lptr, lend, lnew,  # output vars
                                near, foll, dist,
                                ier)  # output var
                self.grid_neighbours = [None for i in range(N)]
                self.grid_neighbours_set = [None for i in range(N)]
                rN = range(N)
                for i in rN:
                    nbsi = []
                    ptr0 = ptr = lend[i]-1
                    for j in rN:
                        nbsi.append(list_[ptr]-1)
                        ptr = lptr[ptr]-1
                        if ptr == ptr0:
                            break
                    self.grid_neighbours[i] = nbsi
                    self.grid_neighbours_set[i] = set(nbsi)
        else:
            raise NotImplementedError("Not yet implemented for \
                                      lat-lon-regular grids!")

        remaining = nodes_set.copy()
        boundary = []
        shape = []
        fullshape = []
        representative = []
        # find a node on the boundary and an outer neighbour:
        lam = 0.5 + gap/2
        lam1 = 1-lam
        while remaining:
            i = list(remaining)[0]
            this_remove = [i]
            cont = False
            while self.grid_neighbours_set[i] <= nodes_set:
                i = self.grid_neighbours[i][int(np.floor(
                    len(self.grid_neighbours[i])*random.uniform()))]
                if i not in remaining:  # we had this earlier
                    cont = True
                    break
                this_remove.append(i)
            remaining -= set(this_remove)
            # if len(nodes_set)==151: print(i,this_remove,remaining,cont)
            if cont:
                continue
            o = list(self.grid_neighbours_set[i] - nodes_set)[0]

            # traverse boundary:
            partial_boundary = [i]
            partial_shape = [lam*pos[i] + lam1*pos[o]]
            partial_fullshape = [0.49*pos[i] + 0.51*pos[o]]
            print(partial_shape)
            steps = [(i, o)]
            for it in range(N):  # at most this many steps we need
                nbi = self.grid_neighbours[i]
                j = nbi[0]
                try:
                    j = nbi[(nbi.index(o)-1) % len(nbi)]
                except IndexError:
                    print("O!", i, o, j, nbi, self.grid_neighbours[o], steps)
                    raise
                if j in nodes_set:
                    i = j
                    partial_boundary.append(i)
                    try:
                        remaining.remove(i)
                    except KeyError:
                        pass
                else:
                    partial_fullshape.append(
                        0.32*pos[i]+0.34*pos[o]+0.34*pos[j])
                    o = j
                partial_shape.append(lam*pos[i] + lam1*pos[o])
                partial_fullshape.append(0.49*pos[i] + 0.51*pos[o])
                if (i, o) in steps:
                    break
                steps.append((i, o))

            mind2 = np.inf
            latlon_shape = []
            latlon_fullshape = []
            length = len(partial_shape)-1
            off = length/2
            for it in range(length):
                pos1 = partial_shape[it]
                pos2 = partial_shape[int((it+off) % length)]
                latlon_shape.append(self.cartesian2latlon(pos1))
                d2 = ((pos2-pos1)**2).sum()
                if d2 < mind2:
                    rep = self.cartesian2latlon((pos1+pos2)/2)
                    mind2 = d2
            latlon_shape.append(self.cartesian2latlon(partial_shape[-1]))
            for pos1 in partial_fullshape:
                latlon_fullshape.append(self.cartesian2latlon(pos1))

            boundary.append(partial_boundary)
            shape.append(latlon_shape)
            fullshape.append(latlon_fullshape)
            representative.append(rep)

        # TODO: sort sub-regions by descending size!
        return boundary, shape, fullshape, representative



# Copyright (C) 2014 SWIPO Project
#
# Authors (this file):
#   Stefan Schinkel <stefan.schinkel@gmail.com>
"""
Provides simple type check for resistive networks
"""

import numpy as np

from pyunicorn import ResNetwork


res = ResNetwork.SmallTestNetwork()


def testAdmittiveDegreeType():
    print("testing types")
    assert isinstance(res.admittive_degree(), np.ndarray)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"
"""
Simple tests for the ClimateNetwork class.
"""
import numpy as np

from pyunicorn.climate.climate_network import ClimateNetwork


def test_str(capsys):
    print(ClimateNetwork.SmallTestNetwork())
    out, err = capsys.readouterr()
    out_ref = "ClimateNetwork:\n" + \
              "GeoNetwork:\n" + \
              "SpatialNetwork:\n" + \
              "Network: undirected, 6 nodes, 7 links, link density 0.467." + \
              "\nGeographical boundaries:\n" + \
              "         time     lat     lon\n" + \
              "   min    0.0    0.00    2.50\n" + \
              "   max    9.0   25.00   15.00\n" + \
              "Threshold: 0.5\n" + \
              "Local connections filtered out: False\n"
    assert out == out_ref


def test_SmallTestNetwork():
    res = ClimateNetwork.SmallTestNetwork().adjacency
    exp = np.array([[0, 0, 0, 1, 1, 1], [0, 0, 1, 1, 1, 0],
                    [0, 1, 0, 0, 1, 0], [1, 1, 0, 0, 0, 0],
                    [1, 1, 1, 0, 0, 0], [1, 0, 0, 0, 0, 0]])
    assert np.allclose(res, exp, atol=1e-04)


def test_link_density_function():
    res = ClimateNetwork.SmallTestNetwork().link_density_function(5)
    exp = (np.array([0., 0.2778, 0.4444, 0.6111, 0.7222]),
           np.array([0.1, 0.28, 0.46, 0.64, 0.82, 1.]))

    assert np.allclose(res[0], exp[0], atol=1e-04)
    assert np.allclose(res[1], exp[1], atol=1e-04)


def test_threshold_from_link_density():
    res = ClimateNetwork.SmallTestNetwork().\
        threshold_from_link_density(link_density=0.5)
    exp = 0.4
    assert np.isclose(res, exp, atol=1e-04)


def test_similarity_measure():
    res = ClimateNetwork.SmallTestNetwork().similarity_measure()[0, :]
    exp = np.array([1., 0.1, 0.2, 0.6, 0.7, 0.55])
    assert np.allclose(res, exp, atol=1e-04)


def test_non_local():
    res = ClimateNetwork.SmallTestNetwork().non_local()
    exp = False
    assert res == exp


def test_set_non_local():
    net = ClimateNetwork.SmallTestNetwork()
    net.set_non_local(non_local=True)

    res = net.adjacency
    exp = np.array([[0, 0, 0, 1, 1, 1], [0, 0, 0, 1, 1, 0],
                    [0, 0, 0, 0, 1, 0], [1, 1, 0, 0, 0, 0],
                    [1, 1, 1, 0, 0, 0], [1, 0, 0, 0, 0, 0]])
    assert np.allclose(res, exp, atol=1e-04)


def test_threshold():
    res = ClimateNetwork.SmallTestNetwork().threshold()
    exp = 0.5
    assert np.isclose(res, exp, atol=1e-04)


def test_set_threshold():
    net = ClimateNetwork.SmallTestNetwork()

    res = net.n_links
    exp = 7
    assert res == exp

    net.set_threshold(threshold=0.7)
    res = net.n_links
    exp = 3
    assert res == exp


def test_set_link_density():
    net = ClimateNetwork.SmallTestNetwork()

    res = net.link_density
    exp = 0.4667
    assert np.isclose(res, exp, atol=1e-04)

    net.set_link_density(link_density=0.7)
    res = net.link_density
    exp = 0.6667
    assert np.isclose(res, exp, atol=1e-04)


def test_correlation_distance():
    res = ClimateNetwork.SmallTestNetwork().correlation_distance().round(2)
    exp = np.array([[0., 0.01, 0.04, 0.18, 0.27, 0.27],
                    [0.01, 0., 0.05, 0.18, 0.29, 0.12],
                    [0.04, 0.05, 0., 0.02, 0.16, 0.03],
                    [0.18, 0.18, 0.01, 0., 0.01, 0.06],
                    [0.27, 0.29, 0.16, 0.01, 0., 0.04],
                    [0.27, 0.12, 0.03, 0.06, 0.04, 0.]])
    assert np.allclose(res, exp, atol=1e-04)


def test_correlation_distance_weighted_closeness():
    res = ClimateNetwork.SmallTestNetwork().\
        correlation_distance_weighted_closeness()
    exp = np.array([0.1646, 0.1351, 0.0894, 0.1096, 0.1659, 0.1102])
    assert np.allclose(res, exp, atol=1e-04)


def test_local_correlation_distance_weighted_vulnerability():
    res = ClimateNetwork.SmallTestNetwork().\
        local_correlation_distance_weighted_vulnerability()
    exp = np.array([0.4037, 0.035, -0.1731, -0.081, 0.3121, -0.0533])
    assert np.allclose(res, exp, atol=1e-04)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Consistency checks for the method caching mix-in.
"""

import pytest
import numpy as np

from pyunicorn.core.cache import Cached


# pylint: disable=disallowed-name
class TestCached:

    class Foo(Cached):
        silence_level = 1

        def __cache_state__(self):
            return ()

        @Cached.method()
        def foo1(self, a: int):
            """Foo1"""
            return a

        @Cached.method(name="foo2")
        def foo2(self, *args):
            """Foo2"""
            return sum(args, start=0)

        @Cached.method()
        def bar(self):
            """Bar"""
            return True

    class Bar(Cached):
        def __init__(self):
            self.counter = 0

        def __cache_state__(self):
            return (self.counter,)

    @classmethod
    def test_args(cls, capfd: pytest.CaptureFixture):
        """
        Dependence on method arguments.
        """
        # immutable instances
        X = cls.Foo()

        # wrapped method metadata
        methods = ["foo1", "foo2", "bar"]
        assert all(getattr(X, f).__doc__ == f.capitalize() for f in methods)

        for _ in range(3):
            # method calls
            X.cache_clear()
            m = Cached.lru_params["maxsize"]

            r1, o1 = [], []
            for _ in range(m + 1):
                r1.append(X.foo1(1))
                o1.append(cls.Foo.foo1.__wrapped__(X, 1))
            r1.append(X.foo1(1.0))
            assert all(r == r1[0] for r in r1)
            assert all(r == o for r, o in zip(r1, o1))

            r2, o2 = [], []
            for i in range(2 * m):
                r2.append(X.foo2(*range(i)))
                o2.append(cls.Foo.foo2.__wrapped__(X, *range(i)))
            assert (np.diff(r2) == range(2 * m - 1)).all()
            assert all(r == o for r, o in zip(r2, o2))

            r3, o3 = [], []
            for i in range(5):
                r3.append(X.bar())
                o3.append(cls.Foo.bar.__wrapped__(X))
            assert all(r and isinstance(r, bool) for r in r3)
            assert all(r == o for r, o in zip(r3, o3))

            # cache lookups
            c1, c2, c3 = (getattr(X, f).cache_info() for f in methods)
            assert c1.maxsize == c2.maxsize == m
            assert (c1.currsize, c1.misses, c1.hits) == (2, 2, m)
            assert (c2.currsize, c2.misses, c2.hits) == (m, 2 * m, 0)
            assert (c3.currsize, c3.misses, c3.hits) == (1, 1, 4)

            # cache clearing
            X.cache_clear(prefix="foo")
            c1, c2, c3 = (getattr(X, f).cache_info() for f in methods)
            assert (c1.currsize, c1.misses, c1.hits) == (0, 0, 0)
            assert (c2.currsize, c2.misses, c2.hits) == (0, 0, 0)
            assert (c3.currsize, c3.misses, c3.hits) == (1, 1, 4)

            # logging behaviour
            capture = capfd.readouterr()
            assert capture.err == ""
            assert capture.out.split("\n")[:-1] == [
                "Calculating foo2..." for _ in range(2 * m)]

    @classmethod
    def test_instance_immutable(cls):
        """
        Dependence on immutable instance attributes.
        """
        # immutable instances
        X, Y = cls.Foo(), cls.Foo()
        m = Cached.lru_params["maxsize"]

        # method calls
        for i in range(m):
            X.foo1(i)
            Y.foo1(i)
        Y.foo1(m-1)

        # cache lookups
        x, y = (o.foo1.cache_info() for o in [X, Y])
        assert x == y
        assert (x.currsize, x.misses, x.hits) == (m, 2 * m, 1)

        # cache clearing
        X.cache_clear()
        x, y = (o.foo1.cache_info() for o in [X, Y])
        assert x == y
        assert (x.currsize, x.misses, x.hits) == (0, 0, 0)

    @classmethod
    def test_instance_mutable(cls):
        """
        Dependence on mutable instance attributes.
        """
        # mutable instances
        class Baz(cls.Bar):
            @Cached.method()
            def baz(self):
                self.counter += 1
                return self.counter

        X, Y = Baz(), Baz()

        # method calls
        m = Cached.lru_params["maxsize"]
        k, n = 3, m // 2
        assert k < n < m
        rs = []
        for i in range(n):
            assert X.baz() == Y.baz()
        Y.counter = 0
        for _ in range(k):
            Y.baz()

        # cache lookups
        x, y = (o.baz.cache_info() for o in [X, Y])
        assert x == y
        assert (x.currsize, x.misses, x.hits) == (2 * n, 2 * n, k)

        # cache clearing
        X.cache_clear()
        x, y = (o.baz.cache_info() for o in [X, Y])
        assert x == y
        assert (x.currsize, x.misses, x.hits) == (0, 0, 0)

    @classmethod
    def test_instance_rec(cls):
        """
        Dependence on owned `Cached` instances.
        """
        # mutable instances
        class BarFoo(cls.Bar):
            def __init__(self, foo: cls.Foo):
                self.foo = foo
                cls.Bar.__init__(self)

            def __rec_cache_state__(self):
                return (self.foo,)

            @Cached.method(attrs=("foo",))
            def baz(self, a: int):
                f = self.foo.foo1(a)
                self.counter += 1
                return f

        # method calls
        m = Cached.lru_params["maxsize"]
        X = BarFoo(cls.Foo())
        for i in range(m):
            assert X.baz(i) == i
        assert X.counter == m
        for i in range(m):
            assert X.baz(i) == i
        assert X.counter == 2 * m

        # cache lookups
        x, y = (m.cache_info() for m in [X.baz, X.foo.foo1])
        assert (x.currsize, x.misses, x.hits) == (m, 2 * m, 0)
        assert (y.currsize, y.misses, y.hits) == (m, m, m)

        # cache clearing
        X.cache_clear()
        x, y = (m.cache_info() for m in [X.baz, X.foo.foo1])
        assert (x.currsize, x.misses, x.hits) == (0, 0, 0)
        assert (y.currsize, y.misses, y.hits) == (0, 0, 0)

    @classmethod
    def test_attributes(cls):
        """
        Dependence on method-specific attributes.
        """
        # mutable instances
        class FooBaz(cls.Foo):
            def __init__(self):
                self.secret = 0

            @Cached.method(attrs=("secret",))
            def baz(self):
                """FooBaz"""
                self.secret += 1
                return self.secret

        class BarBaz(cls.Bar):
            def __init__(self):
                cls.Bar.__init__(self)
                self.secret = 0

            @Cached.method(attrs=("secret",))
            def baz(self):
                """BarBaz"""
                self.counter += 1
                self.secret += 1
                return self.counter + self.secret

        X, Y = FooBaz(), BarBaz()

        # wrapped method metadata
        assert all(o.baz.__doc__ == type(o).__name__ for o in [X, Y])

        # method calls
        m = Cached.lru_params["maxsize"]
        k, n = 3, m // 2
        assert k < n < m
        for _ in range(n):
            X.baz()
            Y.baz()
        print()
        X.secret = 0
        Y.secret = 0
        for _ in range(n):
            X.baz()
            Y.baz()
        print()
        X.secret = 0
        Y.secret = 0
        Y.counter = 0
        for _ in range(k):
            X.baz()
            Y.baz()

        # cache lookups
        x1, y1 = (o.baz.cache_info() for o in [X, Y])
        assert (x1.currsize, x1.misses, x1.hits) == (n, n, n + k)
        assert (y1.currsize, y1.misses, y1.hits) == (2 * n, 2 * n, k)

        # cache clearing
        X.cache_clear()
        x2, y2 = (o.baz.cache_info() for o in [X, Y])
        assert (x2.currsize, x2.misses, x2.hits) == (0, 0, 0)
        assert y2 == y1
        Y.cache_clear()
        x3, y3 = (o.baz.cache_info() for o in [X, Y])
        assert x3 == x2
        assert (y3.currsize, y3.misses, y3.hits) == (0, 0, 0)

    @classmethod
    def test_disable(cls):
        """
        Dependence on global switch.
        """
        Cached.cache_enable = False

        class Baz(cls.Bar):
            def __init__(self):
                cls.Bar.__init__(self)
                self.undeclared_counter = 0

            @Cached.method()
            def baz(self):
                """Baz"""
                self.undeclared_counter += 1

        Cached.cache_enable = True
        X = Baz()

        # wrapped method metadata
        assert X.baz.__doc__ == "Baz"

        # method calls
        m = Cached.lru_params["maxsize"]
        for _ in range(2 * m):
            X.baz()

        # no caching
        assert X.counter == 0
        assert X.undeclared_counter == 2 * m
        assert not hasattr(X.baz, "cache_info")

        # no-op
        X.cache_clear()



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for saving and loading NetCDF files from and to
appropriate Python dictionaries, allowing NetCDF4 compression methods.
"""

#
#  Imports
#

import numpy as np

try:
    from h5netcdf.legacyapi import Dataset
except ImportError:
    try:
        from netCDF4 import Dataset
    except ImportError:
        print("pyunicorn: Packages netCDF4 or h5netcdf could not be loaded. "
              "Some functionality in class NetCDFDictionary might not be "
              "available!")


#
#  Define class NetCDF
#

class NetCDFDictionary:

    """
    Encapsulates appropriate dictionary following NetCDF conventions.

    Also contains methods to load data from NetCDF and NetCDF4 files.
    """

    # TODO: implement silence_level consistently
    def __init__(self, data_dict=None, silence_level=0):
        """
        Return a NetCDF object containing an appropriately structured
        dictionary.

        If no data_dict is given, a default quasi-empty dictionary is created.

        :type data_dict: dictionary
        :arg data_dict: Contains data in a structure following NetCDF
            conventions: {"global_attributes": {}, "dimensions": {},
            "variables": {"obs": {"array": (), "dims": (), "attributes": ()}}}

        :type silence_level: int >= 0
        :arg silence_level: The higher, the less progress info is output.
        """
        if data_dict is None:
            data_dict = {
                "global_attributes": {
                    "title": "Quasi-empty default dictionary"},
                "dimensions": {"x": 1},
                "variables": {
                    "obs": {"array": np.array((1,)), "dims": ('x',),
                            "attributes": {"long_name": "observable"}}}}
        self.dict = data_dict

        self.silence_level = silence_level
        """(int >= 0)
        The higher, the less progress info is output.
        """

    def __str__(self):
        """
        Return a string representation of the object.
        """
        text = (f'NetCDFDictionary:\nGlobal attributes:\n'
                f'{self.dict["global_attributes"]}\nVariables:')

        for key in self.dict["variables"].keys():
            text += (f'\n\t{key}\t-> array shape'
                     f'{self.dict["variables"][key]["array"].shape}')
        return text

    #
    #  Define methods for NetCDF4 files via NetCDF4 module
    #

    @staticmethod
    def from_file(file_name, with_array='all'):
        """
        Load NetCDF4 file into a dictionary.

        Supported file types ``file_type`` are:
          - "NetCDF"
          - "NetCDF4"

        :arg str file_name: The name of the data file.
        :arg [str] with_array: Names of data arrays to be loaded completely.
        :rtype: NetCDF instance
        """
        #  Open NetCDF4 file
        try:
            cdf = Dataset(file_name, "r")
            print(f"MODULE: File {file_name} opened.")
        except RuntimeError:
            print(f"MODULE: File {file_name} couldn't be opened.")
            return None

        #  Create empty dictionary structure
        content = {"global_attributes": {}, "dimensions": {}, "variables": {}}
        #  Copy all global attributes and all dimensions
        content["global_attributes"] = cdf.__dict__
        for dim_name, dim_obj in cdf.dimensions.iteritems():
            content["dimensions"][dim_name] = len(cdf.dimensions[dim_name])

        #  Loop over variables
        for var in cdf.variables.keys():
            #  Create empty dictionary for variable var
            content["variables"][var] = {"array": {}, "type": {}, "dims": {},
                                         "attributes": {}}
            #  Copy type, dimensions and variable attributes
            content["variables"][var]["type"] = cdf.variables[var].dtype.char
            content["variables"][var]["dims"] = cdf.variables[var].dimensions
            content["variables"][var]["attributes"] = \
                cdf.variables[var].__dict__

            #  Load data if wanted
            if var in with_array or with_array == 'all':
                try:
                    content["variables"][var]["array"] = cdf.variables[var][:]
                    print(f"MODULE: Array {var} loaded to dictionary.")
                except MemoryError:
                    print(f"Memory Error during loading of array {var}")
                except RuntimeError:
                    print(f"Other Error during loading of array {var}")

                try:
                    content["variables"][var]["array"] = \
                        content["variables"][var]["array"].astype('float32')
                    print(f"MODULE: Array {var} converted to 'float32'.")
                except MemoryError:
                    print("MODULE: Memory Error during conversion of "
                          f"array {var}.")
                except RuntimeError:
                    print("MODULE: Other Error during conversion of "
                          f"array {var}.")

                #  If a scale_factor is given in the variable, rescale array
                if "scale_factor" in content["variables"][var]["attributes"]:
                    content["variables"][var]["array"] *= \
                        cdf.variables[var].scale_factor
                    del content["variables"][var]["attributes"]["scale_factor"]
                if "add_offset" in content["variables"][var]["attributes"]:
                    content["variables"][var]["array"] += \
                        cdf.variables[var].add_offset
                    del content["variables"][var]["attributes"]["add_offset"]
                #  Recalculate actual_range
                ar_max = content["variables"][var]["array"].max()
                ar_min = content["variables"][var]["array"].min()
                content["variables"][var]["attributes"]["actual_range"] = \
                    np.array([ar_min, ar_max])

        print("MODULE: Dictionary loaded from NetCDF file.")
        cdf.close()

        return NetCDFDictionary(content)

    #  FIXME: createDimension - length of time variable should be "unlimited"
    #  TODO: Change file_name automatically if file already exists
    def to_file(self, file_name, compress=False, comp_level=6,
                least_significant_digit=10):
        """
        Write NetCDF4 file by using appropriate dictionary.

        :arg str file_name: The name of the data file.
        :arg bool compress: Determines whether the data should be compressed.
        :arg int comp_level: Level of compression, between 0 (no compression,
            fastest) and 9 (strongest compression, slowest).
        :arg int least_significant_digit: Last precise digit.
        """
        #  Check dictionary for empty entries
        for val in self.dict.keys():
            if not self.dict[val]:
                print("MODULE: Entry {val} is empty.")

        print(f"MODULE: If {file_name} already existed, old file will be "
              "overwritten.")
        #  Format can be:
        #  NETCDF3_CLASSIC, NETCDF3_64BIT, NETCDF4_CLASSIC, NETCDF4
        cdf = Dataset(file_name, "w", format="NETCDF4")

        #  Write global attributes
        for val in self.dict["global_attributes"]:
            setattr(cdf, val, self.dict["global_attributes"][val])

        #  Write dimensions with given lengths
        for val in self.dict["dimensions"]:
            if val == "time":
                cdf.createDimension(val, self.dict["dimensions"][val])
            else:
                cdf.createDimension(val, self.dict["dimensions"][val])

        #  Write variables
        for var in self.dict["variables"]:
            #  Check variable dictionary for empty entries
            for key in self.dict["variables"][var].keys():
                if not self.dict["variables"][var][key] and key != "type":
                    print(f"MODULE: Entry {key} in variable {var} is empty.")

            var_type = self.dict["variables"][var]["array"].dtype.char
            try:
                var_ = cdf.createVariable(
                    var, var_type, self.dict["variables"][var]["dims"],
                    zlib=compress, complevel=comp_level,
                    least_significant_digit=least_significant_digit)
            except RuntimeError:
                print(f"MODULE: Couldn't create variable {var} "
                      "in NetCDF file.")

            #  Copy the array
            var_[:] = self.dict["variables"][var]["array"]

            #  Calculate actual_range for variables
            ar_max = self.dict["variables"][var]["array"].max()
            ar_min = self.dict["variables"][var]["array"].min()
            self.dict["variables"][var]["attributes"]["actual_range"] = \
                np.array([ar_min, ar_max])

            #  Write all variable attributes to dictionary
            for att in self.dict["variables"][var]["attributes"]:
                setattr(var_, att,
                        self.dict["variables"][var]["attributes"][att])

        cdf.close()
        print("MODULE: Dictionary saved as NetCDF file {file_name}.")



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for analyzing spatially embedded complex networks, handling
multivariate data and generating time series surrogates.
"""

import numpy as np
import matplotlib.pyplot as plt

try:
    import cartopy.crs as ccrs
    import cartopy.feature as cf
except ImportError:
    print("climate: Package cartopy could not be loaded. Some functionality "
          "in class MapPlot might not be available!")

from ..core import Grid

#
#  Define class MapPlot
#


# pylint: disable=too-few-public-methods
class MapPlot:
    """
    Encapsulates map plotting functions via Cartopy and Matplotlib.
    """

    def __init__(self, grid: Grid, title: str):
        """
        :arg grid: The `Grid` object describing the map data to be plotted.
        :arg str title: The title describing the map data.
        """
        self.grid: Grid = grid
        self.title: str = title

        #
        #  Adjust Cartopy settings, fine tuning can be done externally
        #

        # Specify Coordinate Refference System for Map Projection
        # pylint: disable-next=abstract-class-instantiated
        self.projection = ccrs.PlateCarree()

        # Specify CRS (where data should be plotted)
        # pylint: disable-next=abstract-class-instantiated
        self.crs = ccrs.PlateCarree()

        # get spatial dims
        self.lon = self.grid.convert_lon_coordinates(self.grid.lon_sequence())
        self.lat = self.grid.lat_sequence()
        self.gridsize_lon = len(self.lon)
        self.gridsize_lat = len(self.lat)
        self.lon_min = self.grid.boundaries()["lon_min"]
        self.lon_max = self.grid.boundaries()["lon_max"]
        self.lat_min = self.grid.boundaries()["lat_min"]
        self.lat_max = self.grid.boundaries()["lat_max"]

        # extent of data will also give extent of world map
        self.data_extent = [self.lon_min, self.lon_max,
                            self.lat_min, self.lat_max]

    def plot(self, data: np.ndarray, label: str):
        """
        Plot dataset onto ``self.grid``. A simple setup to get a quick view of
        your data.

        The plot can be customized by calling additional Matplotlib or Cartopy
        methods afterwards. It can then be saved via ``plt.savefig()``.

        :arg ndarray data: The dataset to be plotted on the Grid.
        :arg str label: A name for the dataset to print as label.
        """

        #  Generate figure
        plt.figure()

        # create GeoAxes object
        gax = plt.axes(projection=self.projection)

        # create some standards of plotting that can be adjusted
        # before calling ``generate_cartopy_plot()``
        # adjust size and plot coastlines and borders
        gax.set_extent(self.data_extent, crs=self.crs)
        # ax.set_global()
        gax.add_feature(cf.COASTLINE.with_scale("50m"), lw=0.5)
        gax.add_feature(cf.BORDERS.with_scale("50m"), lw=0.2)

        # Draw gridlines in degrees over map
        gl = gax.gridlines(
            crs=self.crs, draw_labels=True,
            linewidth=.6, color='gray',
            alpha=0.5, linestyle='-.'
        )
        gl.xlabel_style = {"size": 7}
        gl.ylabel_style = {"size": 7}

        # plot data upon map
        plt.tricontourf(self.lon, self.lat, data,
                        extent=self.data_extent, transform=self.crs)
        cbar = plt.colorbar(shrink=0.5)
        cbar.set_label(label, rotation=270)

        # add title
        plt.title(self.title)



from pyunicorn import Network, mpi


def do_one():
    net = Network.BarabasiAlbert(n_nodes=100, n_links_each=10)
    return net.global_clustering()


def master():
    n = 1000
    for i in range(0, n):
        mpi.submit_call("do_one", ())
    s = 0
    for i in range(0, n):
        s += mpi.get_next_result()
    print(s/n)
    mpi.info()

mpi.run()



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Created on 28.05.2010

@author: heitzig-j
"""

#
#  Imports
#

import os
import time

import numpy as np
from numpy import random
import matplotlib as mpl
from mpl_toolkits.basemap import Basemap

from .. import nz_coords


class Navigator(object):
    '''
    An interactive network navigator. EXPERIMENTAL
    '''

    def __init__(self,
                 # the network to navigate
                 network=None,
                 # list mapping (super-)node id to parent id (>id)
                 parent=None,
                 # list of information loss in joining children of a supernode
                 loss=None,
                 label=None,
                 autolabels=True,
                 loc=None,
                 solutions=[6, 6**2, 6**3],
                 map=False,
                 degree=None,
                 clustering=None,
                 centrality=None,
                 betweenness=None,
                 distance=None,
                 link_betweenness=None
                 ):
        '''
        Constructor
        '''
        self.map = map
        if map:
            self.fixed = True
            self.bm = Basemap()
            self.bm.drawcoastlines()
            self.fig = mpl.pyplot.gcf()
            self.ax = mpl.pyplot.gca()
            self.max_radius = 90
        else:
            self.fixed = False
            self.fig = mpl.pyplot.figure()
            self.ax = mpl.pyplot.gca()
            self.ax.set_xlim(-10000, 10000)
            self.ax.set_ylim(-10000, 10000)
            self.max_radius = 1000

        # needed so that static event handlers gets self!
        self.fig.pyunicorn_navigator = self
        self._press_cid = self.fig.canvas.mpl_connect('button_press_event',
                                                      self.on_press)
        self._motion_cid = self.fig.canvas.mpl_connect('motion_notify_event',
                                                       self.on_motion)
        self.dragging = None
        self.lastclick = time.time()-1

        self.loss = loss
        self.label = label
        self.autolabels = autolabels
        self.loc = loc
        self.solutions = solutions
        self.dim = 2

        self.set_degree(degree)
        self.set_clustering(clustering)
        self.set_centrality(centrality)
        self.set_betweenness(betweenness)
        self.set_distance(distance)
        self.set_link_betweenness(link_betweenness)

        self.set_network(network)
        if parent is not None:
            self.set_tree(parent, loss)

    @staticmethod
    def on_press(event):
        if event.inaxes is None:
            return  # not left clicked
        self = event.inaxes.figure.pyunicorn_navigator
        self.dragging = None
        if event.button == 1:
            if time.time() - self.lastclick < 0.3:  # doubleclick
                self._process_expand_button_on(self.get_target(event))
                self.lastclick = time.time()-1
            else:
                self.lastclick = time.time()
        elif event.button == 3:
            self._process_collapse_button_on(self.get_target(event))

    @staticmethod
    def on_motion(event):
        if event.inaxes is None:
            return
        self = event.inaxes.figure.pyunicorn_navigator
        if self.fixed:
            return
        if event.button != 1:
            self.dragging = None
            return  # not dragging
        pos = np.array([[event.xdata, event.ydata]])
        if self.dragging is None:
            dist = np.sqrt(((
                self.position[self.is_shown, :] - pos)**2).sum(axis=1))
            cands = np.where(dist < self.radius[self.is_shown] + 500.0)[0]
            if cands.size == 0:
                return
            self.dragging = np.arange(
                self.S)[self.is_shown][cands[dist[cands].argmin()]]
        self._move(self.dragging, pos.flatten())
        # self._improve_step()

        # TODO: on hover, show controls!

    def get_target(self, event):
        pos = np.array([[event.xdata, event.ydata]])
        dist = np.sqrt(((
            self.position[self.is_shown, :] - pos)**2).sum(axis=1))
        cands = np.where(dist < self.max_radius)[0]
        if cands.size == 0:
            return None
        return np.arange(self.S)[self.is_shown][cands[dist[cands].argmin()]]

    def set_network(self, network):
        self.network = network
        if network is not None and "parent" in self.__dict__:
            self._process_network_and_tree()
        # TODO: adjust

    def set_tree(self, parent, loss=None):
        self.parent = parent.tolist()
        self.loss = loss
        S = self.S = len(parent)+1
        # expansion state of supernodes by id
        self.is_expanded = np.zeros(S).astype("bool")
        # whether supernode is shown
        self.is_shown = np.zeros(S).astype("bool")
        self.shown = set([])
        # position of (shown) supernodes
        self.position = np.zeros((S, self.dim))
        self.lines = [{} for i in range(S)]
        self.circles = [None for i in range(S)]
        self.polys = [[] for i in range(S)]
        self.children = [[] for i in range(S)]
        for i in range(S-1):
            if parent[i] >= 0:
                self.children[parent[i]].append(i)
        if self.network is not None:
            self._process_network_and_tree()

    def set_degree(self, k):
        self.k0 = k

    def set_clustering(self, C):
        self.C0 = C

    def set_centrality(self, CC):
        self.CC0 = CC

    def set_betweenness(self, B):
        self.B0 = B

    def set_distance(self, D):
        self.D0 = D

    def set_link_betweenness(self, LB):
        self.LB0 = LB

    def _process_network_and_tree(self):
        # determine weights and linked proportions:
        N = self.N = self.network.N
        S = self.S
        self.weight = np.zeros(self.S)  # weight of supernodes
        sum_k = np.zeros(S)
        sum_C = np.zeros(S)
        sum_CC = np.zeros(S)
        sum_B = np.zeros(S)
        self.weight[:self.N] = \
            self.network.node_weights / self.network.total_node_weight
        if self.k0 is None:
            sum_k[:N] = np.zeros(N)
        else:
            sum_k[:N] = self.k0
        sum_k *= self.weight
        if self.C0 is None:
            sum_C[:N] = np.zeros(N)
        else:
            sum_C[:N] = self.C0
        sum_C *= self.weight
        if self.CC0 is None:
            sum_CC[:N] = np.zeros(N)
        else:
            sum_CC[:N] = self.CC0
        sum_CC *= self.weight
        if self.B0 is None:
            sum_B[:N] = np.zeros(N)
        else:
            sum_B[:N] = self.B0
        sum_B *= self.weight
        self.linked_weight = np.zeros((S, S))  # csc_matrix((self.S,self.S))
        self.linked_weight[range(N), range(N)] = self.weight[:self.N]**2
        for i, j in nz_coords(self.network.sp_A):
            self.linked_weight[i, j] = self.weight[i] * self.weight[j]
        if self.D0 is not None:
            sum_D = np.zeros((S, S))
            sum_D[:N, :N] = self.D0 * (self.weight[:N].reshape((-1, 1)) *
                                       self.weight[:N].reshape((1, -1)))
            print(sum_D)
        else:
            sum_D = None
        if self.LB0 is not None:
            sum_LB = np.zeros((S, S))
            sum_LB[:N, :N] = self.LB0 * self.linked_weight[:N, :N]
        else:
            sum_LB = None
        self.nodes = [[i] for i in range(N)] + [[] for i in range(N, S)]
        # this relies on parents having higher indices:
        for i in range(self.N, self.S):
            C = self.children[i]
            self.weight[i] = self.weight[C].sum()
            sum_k[i] = sum_k[C].sum()
            sum_C[i] = sum_C[C].sum()
            sum_CC[i] = sum_CC[C].sum()
            sum_B[i] = sum_B[C].sum()
            lwi = np.zeros(S)
            for j in C:  # TODO!!
                self.nodes[i] += self.nodes[j]
                lwi += self.linked_weight[:, j]
            self.linked_weight[:, i] = self.linked_weight[i, :] = lwi
            if sum_D is not None:
                di = np.zeros(S)
                for j in C:
                    di += sum_D[:, j]
                sum_D[:, i] = sum_D[i, :] = di
            if sum_LB is not None:
                lbi = np.zeros(S)
                for j in C:
                    lbi += sum_LB[:, j]
                sum_LB[:, i] = sum_LB[i, :] = lbi
        self.k = sum_k / self.weight
        self.C = sum_C / self.weight
        self.CC = sum_CC / self.weight
        self.B = sum_B / self.weight
        fac = 1 / (self.weight.reshape((-1, 1)) * self.weight.reshape((1, -1)))
        self.linked_proportion = self.linked_weight * fac
        if sum_D is not None:
            self.D = sum_D * fac
            print(sum_D)
            print(self.D)
        if sum_LB is not None:
            self.LB = sum_LB * fac
        del fac
        uc = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        lc = "abcdefghijklmnopqrstuvwxyz"
        labelparts = [[uc[i] for i in range(23)]
                      + ["X"+uc[i] for i in range(26)]
                      + ["Y"+uc[i] for i in range(26)]
                      + ["Z"+uc[i] for i in range(26)],
                      range(1, 100),
                      [lc[i] for i in range(23)]
                      + ["x"+lc[i] for i in range(26)]
                      + ["y"+lc[i] for i in range(26)]
                      + ["z"+lc[i] for i in range(26)],
                      range(1, 100),
                      [lc[i] for i in range(23)]
                      + ["x"+lc[i] for i in range(26)]
                      + ["y"+lc[i] for i in range(26)]
                      + ["z"+lc[i] for i in range(26)]]
        labelnullpart = "~~~~~"
        if self.autolabels:
            self.label = [str(S-i) for i in range(S)]
            self.label[S-1] = ""
            pa = [S-1]
            grandpa = np.zeros(S).astype("int") + S-1
            self.solution_nodes = [None for s in self.solutions]
            level = -1
            while len(pa) < np.max(self.solutions):
                c = max(pa)
                pa.remove(c)
                C = self.children[c]
                pa += C
                grandpa[C] = grandpa[c]
                if len(pa) in self.solutions:
                    level += 1
                    self.solution_nodes[level] = pa+[]
                    paas = (-self.weight[pa]).argsort()
                    inds = np.zeros(S).astype("int")
                    for i in range(len(pa)):
                        j = pa[paas[i]]
                        gp = grandpa[j]
                        if gp != j:
                            self.label[j] = self.label[gp] + \
                                str(labelparts[level][inds[gp]])
                            inds[gp] += 1
                        else:
                            self.label[j] = self.label[gp] + \
                                labelnullpart[level]
                        print(j, self.label[j])
                    grandpa[pa] = pa
                elif level >= 0:
                    for i in C:
                        self.label[i] = self.label[grandpa[i]] + "." + str(S-i)
                        print(i, self.label[i])
        elif self.label is None:
            self.label = [str(i) for i in range(S)]
        self.is_geo = "grid" in self.network.__dict__
        if self.is_geo:
            self.lat = np.zeros(S)
            self.lon = np.zeros(S)
            self.lat[:N] = self.network.grid.get_lat_sequence()
            self.lon[:N] = self.network.grid.get_lon_sequence()
        self.boundary = [[] for i in range(S)]
        self.shape = [[] for i in range(S)]
        self.fullshape = [[] for i in range(S)]
        self.representative = [None for i in range(S)]
        if self.is_geo:
            for i in range(S):
                if self.label != str(S-i):
                    (
                        self.boundary[i], self.shape[i], self.fullshape[i],
                        self.representative[i]
                    ) = self.network.get_boundary(self.nodes[i], gap=0.1)
                    self.representative[i].append((0, 0))
                    self.lat[i], self.lon[i] = self.representative[i][0]
        if self.map:
            for i in range(self.S):
                self.position[i, 0], self.position[i, 1] = \
                    self.bm(self.lon[i], self.lat[i])
            #  radius of (shown) supernodes
            self.radius = 0.2*self.max_radius * \
                np.sqrt(self.weight/self.weight.max())
        else:
            # radius of (shown) supernodes
            self.radius = self.max_radius * \
                np.sqrt(self.weight/self.weight.max())
        color = random.uniform(size=(S, 3))
        color = np.floor(
            64+191*color/color.sum(axis=1).reshape((S, 1))).astype("int")
        self.color = \
            ["#"+hex(color[i, 0]*256**2+color[i, 1]*256+color[i, 2])[2:8]
             for i in range(S)]
        # TODO: use given distances!
        self.preferred_distance = 2000 * (10 - 9*self.linked_weight)  # 2000
        # self._process_expand_button_on(self.S-1)

    def draw_field(self, values, filename=None):
        fig = mpl.pyplot.figure()
        # fig.set_size_inches(18,9)
        ax = fig.add_axes((0, 0, 1, 1))
        bm = Basemap(ax=ax)
        bm.drawcoastlines(ax=ax)
        # fig.savefig("test.png",
        #             bbox_inches=mpl.transforms.Bbox([[0,0],[18,9]]))
        # fig.savefig("test.png",
        #             bbox_inches=mpl.transforms.Bbox([[0,1.03125],
        #                                              [8.125,5.09375]]))
        vmin = np.min(values)
        vrange = np.max(values)-vmin
        for i in range(self.N):
            b = (values[i]-vmin)*2.0/vrange
            if b < 1:
                c = [b, b, 1.0]
            else:
                c = [1.0, 2.0-b, 2.0-b]
            for sh in self.fullshape[i]:
                sh = np.array(sh)
                x, y = bm(sh[:, 1], sh[:, 0])
                xy = np.zeros((x.size, 2))
                xy[:, 0] = x
                xy[:, 1] = y
                if x.max()-x.min() < 180:
                    ax.add_patch(mpl.patches.Polygon(
                        xy, fc=c, lw=0.0, fill=True, ec="none", aa=False))
                else:
                    x1 = 1.0*x
                    x1[np.where(x < 0)] += 360.0
                    xy[:, 0] = x1
                    ax.add_patch(mpl.patches.Polygon(
                        xy, fc=c, lw=0.0, fill=True, ec="none", aa=False))
                    x1 -= 360.0
                    xy1 = 1*xy
                    xy1[:, 0] = x1
                    ax.add_patch(mpl.patches.Polygon(
                        xy1, fc=c, lw=0.0, fill=True, ec="none", aa=False))
        fig.show()
        if filename is not None:
            # fig.savefig(filename, bbox_inches=mpl.transforms.Bbox(
            #     [[0.02, 1.05], [8.12, 5.09]]))
            fig.savefig(filename, bbox_inches=mpl.transforms.Bbox(
                [[0.02, 1.0], [7.99, 5.0]]))
        return fig

    def write_kml(
      self,
      path="climate-network-analysis",
      pathurl="http://localhost/~heitzig-j/climate-network-analysis",
      title="Climate Network Analysis",
      measures=None,
      stats=None):
        # TODO: kmz, betweenness correctly, i.png, zip, doublettes
        f = open(path+"/open-this-in-googleearth.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
  <NetworkLink>
    <name>&lt;i&gt;&lt;b&gt;"""+title+"""&lt;/b&gt;&lt;/i&gt;</name>
    <Snippet maxLines="1">(please expand folders as needed)</Snippet>
    <open>1</open>
    <Link>
<!--
*******************************************************************************
** if you unzipped the archive into your own (local) webserver,              **
** change the following to point to the file doc.kml inside it,              **
** for example:                                                              **
**  <href>http://localhost/~username/climate-network-analysis/doc.kml</href> **
*******************************************************************************
-->
      <href>"""+pathurl+"""/doc.kml</href>
    </Link>
  </NetworkLink>
</kml>""")
        f.close()
        f = open(path+"/standard.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>doc.kml</targetHref>
    <Change>
      <Folder targetId="resolution"><visibility>1</visibility></Folder>
      <Folder targetId="overlays"><visibility>0</visibility></Folder>
      <Folder targetId="options"><visibility>1</visibility></Folder>
      <Folder targetId="nodes"><visibility>1</visibility></Folder>
      <Folder targetId="links"><visibility>1</visibility></Folder>
      <Folder targetId="labels"><visibility>1</visibility></Folder>
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        f = open(path+"/no-standard.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>doc.kml</targetHref>
    <Change>
      <Folder targetId="standard"><visibility>0</visibility></Folder>
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        # TODO: tour
        f = open(path+"/doc.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2" """
                """xmlns:gx="http://www.google.com/kml/ext/2.2">
  <Document>
    <Style id="check">
      <ListStyle><listItemType>check</listItemType></ListStyle>
    </Style>
    <Style id="radio">
      <ListStyle><listItemType>radioFolder</listItemType></ListStyle>
    </Style>
    <Style id="hide">
      <ListStyle><listItemType>checkHideChildren</listItemType><ItemIcon/>"""
                """</ListStyle>
    </Style>
    <Style id="noicon">
      <ListStyle><ItemIcon/></ListStyle>
    </Style>
    <!--<gx:Tour id="intro">
      <name>&lt;a href=&quot;#intro;flyto&quot;&gt;Introductory tour&lt;"""
                """/a&gt;</name>
      <gx:Playlist>
        <gx:AnimatedUpdate>
          <Update>
          </Update>
        </gx:AnimatedUpdate>
      </gx:Playlist>
    </gx:Tour>-->
    <Folder id="standard">
      <name>&lt;b&gt;CHECK THIS FOR STANDARD SETTINGS&lt;/b&gt;</name>
      <visibility>0</visibility><styleUrl>#hide</styleUrl>
      <NetworkLink><Link><href>standard.kml</href></Link></NetworkLink>
    </Folder>
    <Folder id="resolution">
      <name>&lt;b&gt;Network resolution:&lt;/b&gt;</name>
      <open>1</open><visibility>1</visibility><styleUrl>#radio</styleUrl>
      <Folder>
        <name>Hierarchy of regions (with slider)...</name>
        <visibility>0</visibility><styleUrl>#radio</styleUrl>
        <Folder>
          <name>&lt;i&gt;show regions only&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#hide</styleUrl>
          <NetworkLink><Link><href>links-off.kml</href></Link></NetworkLink>
          <NetworkLink><Link><href>hierarchy-regions.kml</href></Link>"""
                """</NetworkLink>
        </Folder>
        <Folder>
          <name>&lt;i&gt;regions and links&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#hide</styleUrl>
          <NetworkLink><Link><href>no-standard.kml</href></Link></NetworkLink>
          <NetworkLink><Link><href>hierarchy-regions.kml</href></Link>"""
                """</NetworkLink>
          <NetworkLink><Link><href>hierarchy-links.kml</href></Link>"""
                """</NetworkLink>
        </Folder>
        <Folder>
          <name>&lt;i&gt;links only&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#hide</styleUrl>
          <NetworkLink><Link><href>no-standard.kml</href></Link></NetworkLink>
          <NetworkLink><Link><href>regions-off.kml</href></Link></NetworkLink>
          <NetworkLink><Link><href>hierarchy-links.kml</href></Link>"""
                """</NetworkLink>
        </Folder>
      </Folder>""")
        for s in self.solutions:
            f.write("""
      <Folder>
        <name>Partition into """+str(s)+""" regions...</name>
        <visibility>0</visibility><styleUrl>#radio</styleUrl>
        <Folder>
          <name>&lt;i&gt;show regions only&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#hide</styleUrl>
          <NetworkLink><Link><href>no-standard.kml</href></Link></NetworkLink>
          <NetworkLink><Link><href>links-off.kml</href></Link></NetworkLink>
          <NetworkLink><Link><href>partition"""+str(s)
                    + """-regions.kml</href></Link></NetworkLink>
        </Folder>
        <Folder>
          <name>&lt;i&gt;regions and links&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#hide</styleUrl>
          <NetworkLink><Link><href>no-standard.kml</href></Link></NetworkLink>
          <NetworkLink><Link><href>partition"""+str(s)
                    + """-regions.kml</href></Link></NetworkLink>
          <NetworkLink><Link><href>partition"""+str(s)
                    + """-links.kml</href></Link></NetworkLink>
        </Folder>
        <Folder>
          <name>&lt;i&gt;links only&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#hide</styleUrl>
          <NetworkLink><Link><href>no-standard.kml</href></Link></NetworkLink>
          <NetworkLink><Link><href>regions-off.kml</href></Link></NetworkLink>
          <NetworkLink><Link><href>partition"""+str(s)
                    + """-links.kml</href></Link></NetworkLink>
        </Folder>
      </Folder>""")
        f.write("""
    </Folder>
    <Folder id="overlays">
      <name>&lt;b&gt;Color-coded maps:&lt;/b&gt;</name>
      <open>1</open><visibility>0</visibility><styleUrl>#radio</styleUrl>""")
        f.write("""
      <Folder>
        <name>Network measures...</name>
        <open>0</open><visibility>0</visibility><styleUrl>#radio</styleUrl>""")
        if measures is not None and len(measures) > 0:
            for i in range(len(measures)):
                self.draw_field(measures[i][1],
                                filename=path+"/measure"+str(i)+".png")
                f.write("""
        <GroundOverlay>
          <name>&lt;i&gt;"""+measures[i][0]+"""&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>
          <drawOrder>1</drawOrder>
          <Icon><href>measure"""+str(i)+""".png</href></Icon>
          <Region>
            <LatLonAltBox><north>90</north><south>-90</south><east>180.1"""
                        """</east><west>-180.1</west><minAltitude>0"""
                        """</minAltitude><maxAltitude>0</maxAltitude>"""
                        """</LatLonAltBox>
            <Lod><minLodPixels>5</minLodPixels><maxLodPixels>12000"""
                        """</maxLodPixels><minFadeExtent>5</minFadeExtent>"""
                        """<maxFadeExtent>11000</maxFadeExtent></Lod>
          </Region>
          <LatLonBox><north>90</north><south>-90</south><east>-180.1</east>"""
                        """<west>180.1</west></LatLonBox>
        </GroundOverlay>""")
            f.write("""
      </Folder>""")
        if stats is not None and len(stats) > 0:
            f.write("""
      <Folder>
        <name>Underlying data...</name>
        <open>0</open><visibility>0</visibility><styleUrl>#radio</styleUrl>""")
            for i in range(len(stats)):
                f.write("""
        <GroundOverlay>
          <name>&lt;i&gt;l"""+stats[i][0]+"""&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>
          <drawOrder>1</drawOrder>
          <Icon><href>stat"""+str(0)+""".png</href></Icon>
          <Region>
            <LatLonAltBox><north>90</north><south>-90</south><east>180.1"""
                        """</east><west>-180.1</west><minAltitude>0"""
                        """</minAltitude><maxAltitude>0</maxAltitude>"""
                        """</LatLonAltBox>
            <Lod><minLodPixels>5</minLodPixels><maxLodPixels>12000"""
                        """</maxLodPixels><minFadeExtent>5</minFadeExtent>"""
                        """<maxFadeExtent>11000</maxFadeExtent></Lod>
          </Region>
          <LatLonBox><north>90</north><south>-90</south><east>-180</east>"""
                        """<west>180</west></LatLonBox>
        </GroundOverlay>""")
            f.write("""
      </Folder>""")
        f.write("""
    </Folder>
    <Folder id="options">
      <name>&lt;b&gt;Options:&lt;/b&gt;</name>
      <open>1</open><visibility>0</visibility>
      <Folder id="nodes">
        <name>Node bullets...</name>
        <visibility>0</visibility><styleUrl>#radio</styleUrl>""")
        if self.B0 is not None:
            f.write("""
        <NetworkLink>
          <name>&lt;i&gt;colored by centrality and betweenness&lt;/i&gt;"""
                    """</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>
          <Link><href>nodes-betweenness.kml</href></Link>
        </NetworkLink>
        <NetworkLink>
          <name>&lt;i&gt;yellow&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>""")
        else:
            f.write("""
        <NetworkLink>
          <name>&lt;i&gt;yellow&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>""")
        f.write("""
          <Link><href>nodes-yellow.kml</href></Link>
        </NetworkLink>
        <NetworkLink>
          <name>&lt;i&gt;black&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>
          <Link><href>nodes-black.kml</href></Link>
        </NetworkLink>
        <NetworkLink>
          <name>&lt;i&gt;off&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>
          <Link><href>nodes-off.kml</href></Link>
        </NetworkLink>
      </Folder>
      <Folder id="links">
        <name>Link color...</name>
        <visibility>0</visibility><styleUrl>#radio</styleUrl>""")
        if self.LB0 is not None:
            f.write("""
        <NetworkLink>
          <name>&lt;i&gt;by network distance and betweenness&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>
          <Link><href>links-betweenness.kml</href></Link>
        </NetworkLink>
        <NetworkLink>
          <name>&lt;i&gt;yellow&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>""")
        else:
            f.write("""
        <NetworkLink>
          <name>&lt;i&gt;yellow&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>""")
        f.write("""
          <Link><href>links-yellow.kml</href></Link>
        </NetworkLink>
        <NetworkLink>
          <name>&lt;i&gt;dark blue&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>
          <Link><href>links-black.kml</href></Link>
        </NetworkLink>
      </Folder>
      <Folder id="labels">
        <name>Labels...</name>
        <visibility>0</visibility><styleUrl>#radio</styleUrl>
        <NetworkLink>
          <name>&lt;i&gt;large but simple&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>
          <Link><href>labels-large.kml</href></Link>
        </NetworkLink>
        <NetworkLink>
          <name>&lt;i&gt;small with statistics&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>
          <Link><href>labels-small.kml</href></Link>
        </NetworkLink>
        <NetworkLink>
          <name>&lt;i&gt;off&lt;/i&gt;</name>
          <visibility>0</visibility><styleUrl>#noicon</styleUrl>
          <Link><href>labels-off.kml</href></Link>
        </NetworkLink>
      </Folder>
    </Folder>
    <NetworkLink>
      <name>&lt;small&gt;(don't deselect this)&lt;/small&gt;</name>
      <open>0</open><visibility>1</visibility><refreshVisibility>0"""
                """</refreshVisibility><styleUrl>#hide</styleUrl>
      <Link><href>data.kmz</href></Link>
    </NetworkLink>
  </Document>
</kml>""")
        f.close()
        f = open(path+"/regions-off.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>
      <Style targetId="a"><PolyStyle><outline>0</outline></PolyStyle></Style>
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        f = open(path+"/links-off.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>
      <Folder targetId="links"><visibility>0</visibility></Folder>
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()

        # determine used regions and links:
        self.regions = self.solution_nodes[0]+[]
        pa = self.solution_nodes[0]+[]
        while len(pa) < np.max(self.solutions):
            c = max(pa)
            pa.remove(c)
            C = self.children[c]
            pa += C
            self.regions += C
        links = []
        for i in self.regions:
            for j in self.regions:
                if j < i < self.parent[j] and self.linked_weight[i, j] > 0:
                    links.append((i, j))

        if self.B0 is not None:
            bmax = np.log(0.0001+max(self.B[self.regions]))
            bmin = np.log(0.0001+min(self.B[self.regions]))
            ccmax = max(self.CC[self.regions])
            f = open(path+"/nodes-betweenness.kml", "w")
            f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
            for i in self.regions:
                b = int((np.log(0.0001+self.B[i])-bmin)/(bmax-bmin)*512)
                if b < 256:
                    c = hex(511-b)[-2:] + "ffff"
                else:
                    c = "00" + hex(767-b)[-2:] + "ff"
                a = hex(256+int(self.CC[i]/ccmax*255))[-2:]
                f.write("""
<IconStyle targetId="i"""+str(i)+""""><color>"""+a+c
                        + """</color><scale>1.5</scale></IconStyle>""")
            f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
            f.close()
        f = open(path+"/nodes-yellow.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
        for i in self.regions:
            f.write("""
<IconStyle targetId="i"""+str(i)
                    + """"><color>ff00ffff</color><scale>1.5</scale>"""
                    """</IconStyle>""")
        f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        f = open(path+"/nodes-black.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
        for i in self.regions:
            f.write("""
<IconStyle targetId="i"""+str(i)
                    + """"><color>ff000000</color><scale>1.5</scale>"""
                    """</IconStyle>""")
        f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        f = open(path+"/nodes-off.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
        for i in self.regions:
            f.write("""
<IconStyle targetId="i"""+str(i)+""""><scale>0</scale></IconStyle>""")
        f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        if self.LB0 is not None:
            bmin = np.inf
            bmax = -np.inf
            for i, j in links:
                bmax = max(bmax, self.LB[i, j])
                bmin = min(bmin, self.LB[i, j])
            bmax = np.log(0.0001+bmax)
            bmin = np.log(0.0001+bmin)
            if self.D0 is not None:
                dmax = 0.0
                for i, j in links:
                    dmax = max(dmax, self.D[i, j])
                print("DMAX", dmax)
            f = open(path+"/links-betweenness.kml", "w")
            f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
            for i, j in links:
                b = int((np.log(0.0001+self.LB[i, j])-bmin)/(bmax-bmin)*512)
                if b < 256:
                    c = hex(511-b)[-2:] + "ffff"
                else:
                    c = "00" + hex(767-b)[-2:] + "ff"
                if self.D0 is None:
                    a = "ff"
                elif np.isnan(self.D[i, j]):
                    a = "ff"
                    print(i, j, "nan")
                else:
                    a = hex(511-int((self.D[i, j]-1)/(dmax-1)*255))[-2:]
                f.write("""
<LineStyle targetId="a"""+str(i)+"-"+str(j)+""""><color>"""+a+c
                        + """</color></LineStyle>""")
            f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
            f.close()
        f = open(path+"/links-yellow.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
        for i, j in links:
            f.write("""
<LineStyle targetId="a"""+str(i)+"-"+str(j)
                    + """"><color>ff00ffff</color></LineStyle>""")
        f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        f = open(path+"/links-black.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
        for i, j in links:
            f.write("""
<LineStyle targetId="a"""+str(i)+"-"+str(j)
                    + """"><color>ff7f0000</color></LineStyle>""")
        f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        f = open(path+"/labels-large.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
        for i in self.regions:
            f.write("""
<Placemark targetId="p"""+str(i)+""""><name>"""+self.label[i]
                    + """</name></Placemark><LabelStyle targetId="l"""
                    + str(i)+""""><scale>3</scale></LabelStyle>""")
        for i, j in links:
            f.write("""
<Placemark targetId="p"""+str(i)+"-"+str(j)+""""><name>"""+self.label[i]
                    + " - "+self.label[j]+"""</name></Placemark><LabelStyle"""
                    """ targetId="l"""+str(i)+"-"+str(j)
                    + """"><scale>2</scale></LabelStyle>""")
        f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        f = open(path+"/labels-small.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
        for i in self.regions:
            lab = self.label[i] + \
                (", dg %(k)6.2f, cls %(C)2d, btw %(B)5.2f" % {
                    "k": self.k[i]*510.07,
                    "C": self.C[i]*100,
                    "B": self.B[i]*100}).\
                replace("  ", " ").replace("  ", " ")
            # TODO: use provided data
            f.write("""
<Placemark targetId="p"""+str(i)+""""><name>"""+lab
                    + """</name></Placemark><LabelStyle targetId="l"""+str(i)
                    + """"><scale>3</scale></LabelStyle>""")
        for i, j in links:
            lab = self.label[i] + " - " + self.label[j] + \
                (", str %(s)5.2f, dst %(l)6.2f, btw %(B)7.4f" % {
                    "s": self.linked_weight[i, j]*510.07**2,
                    "l": self.D[i, j],
                    "B": self.LB[i, j]*100}).\
                replace("  ", " ").replace("  ", " ")
            # TODO: use provided data
            f.write("""
<Placemark targetId="p"""+str(i)+"-"+str(j)+""""><name>"""+lab
                    + """</name></Placemark><LabelStyle targetId="l"""+str(i)
                    + "-"+str(j)+""""><scale>2</scale></LabelStyle>""")
        f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        f = open(path+"/labels-off.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
        for i in self.regions:
            f.write("""
<Placemark targetId="p"""+str(i)+""""><name>"""+self.label[i]
                    + """</name></Placemark><LabelStyle targetId="l"""+str(i)
                    + """"><scale>0</scale></LabelStyle>""")
        for i, j in links:
            f.write("""
<Placemark targetId="p"""+str(i)+"-"+str(j)+""""><name>"""+self.label[i]
                    + " - "+self.label[j]+"""</name></Placemark><LabelStyle """
                    """targetId="l"""+str(i)+"-"+str(j)
                    + """"><scale>0</scale></LabelStyle>""")
        f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        f = open(path+"/hierarchy-regions.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>
      <Style targetId="a"><PolyStyle><outline>1</outline>"""
                """</PolyStyle></Style>""")
        for i in self.regions:
            t1 = max(np.min(self.solutions), self.S+1-self.parent[i])
            t2 = min(self.S+1-i, np.max(self.solutions)+1)-1
            f.write("""
<Placemark targetId="p"""+str(i)+""""><TimeSpan><begin>"""+str(t1)
                    + """</begin><end>"""+str(t2)
                    + """-12-31</end></TimeSpan><LookAt><gx:TimeStamp><when>"""
                    + str(t1)
                    + """</when></gx:TimeStamp></LookAt></Placemark>""")
        f.write("""
      <Folder targetId="regions"><visibility>1</visibility></Folder>
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        f = open(path+"/hierarchy-links.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
        for i, j in links:
            t1 = max(np.min(self.solutions), self.S+1-self.parent[i],
                     self.S+1-self.parent[j])
            t2 = min(self.S+1-i, self.S+1-j, np.max(self.solutions)+1)-1
            f.write("""
<Placemark targetId="p"""+str(i)+"-"+str(j)+""""><TimeSpan><begin>"""+str(t1)
                    + """</begin><end>"""+str(t2)
                    + """-12-31</end></TimeSpan></Placemark>""")
        f.write("""
      <Folder targetId="links"><visibility>1</visibility></Folder>
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
        f.close()
        for level, s in enumerate(self.solutions):
            f = open(path+"/partition"+str(s)+"-regions.kml", "w")
            f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>
      <Style targetId="a"><PolyStyle><outline>1</outline>"""
                    """</PolyStyle></Style>""")
            for i in self.solution_nodes[level]:
                f.write("""
<Placemark targetId="p"""+str(i)+""""><TimeSpan><begin/><end/></TimeSpan>"""
                        """<LookAt><gx:TimeStamp><when/></gx:TimeStamp>"""
                        """</LookAt></Placemark>""")
            f.write("""
      <Folder targetId="regions"><visibility>1</visibility></Folder>
      <Folder targetId="regions"""+str(s)
                    + """"><visibility>1</visibility></Folder>
      <Folder targetId="regions0"><visibility>0</visibility></Folder>""")
            for level2, s2 in enumerate(self.solutions):
                if level2 != level:
                    f.write("""
      <Folder targetId="regions"""+str(s2)
                            + """"><visibility>0</visibility></Folder>""")
            f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
            f.close()
            f = open(path+"/partition"+str(s)+"-links.kml", "w")
            f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
<NetworkLinkControl>
  <Update>
    <targetHref>data.kmz</targetHref>
    <Change>""")
            for i in self.solution_nodes[level]:
                for j in self.solution_nodes[level]:
                    if (i, j) in links:
                        f.write("""
<Placemark targetId="p"""+str(i)+"-"+str(j)
                                + """"><TimeSpan><begin/><end/></TimeSpan>"""
                                """</Placemark>""")
            f.write("""
      <Folder targetId="links"><visibility>1</visibility></Folder>
      <Folder targetId="links"""+str(s)
                    + """"><visibility>1</visibility></Folder>
      <Folder targetId="links0"><visibility>0</visibility></Folder>""")
            for level2, s2 in enumerate(self.solutions):
                if level2 != level:
                    f.write("""
      <Folder targetId="links"""+str(s2)
                            + """"><visibility>0</visibility></Folder>""")
            f.write("""
    </Change>
  </Update>
</NetworkLinkControl>
</kml>""")
            f.close()
        f = open(path+"/data.kml", "w")
        f.write("""<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2" """
                """xmlns:gx="http://www.google.com/kml/ext/2.2">
 <Document>
  <Folder><visibility>1</visibility></Folder>
  <Style id="a"><!-- common to regions -->
    <PolyStyle>
      <fill>0</fill><outline>1</outline>
    </PolyStyle>
    <BalloonStyle>
      <text><![CDATA[
        <b>$[l]</b><br/>
        <table>
          <tr><td>Area</td><td>$[a]&nbsp;Mm²</td></tr>
          <tr><td>Ave. degree measure</td><td>$[k]&nbsp;Mm²</td></tr>
          <tr><td>Ave. clustering measure</td><td nowrap>$[C] %</td></tr>
          <tr><td>Ave. centrality measure</td><td nowrap>$[CC] %</td></tr>
          <tr><td>Ave. betweenness measure</td><td nowrap>$[B] %</td></tr>
          <tr><td>contained in region</td><td><a href="#p$[p];balloonFlyto">"""
                """$[pl]</a></td></tr>
          <tr><td>largest contained subregions</td><td><a href="#p$[n1];"""
                """balloonFlyto">$[l1]</a>, <a href="#p$[n2];balloonFlyto">"""
                """$[l2]</a></td></tr>
        </table>
      ]]></text>
    </BalloonStyle>
  </Style>
  <Style id="b"><!-- highlit regions -->
    <PolyStyle>
      <fill>0</fill><outline>1</outline>
    </PolyStyle>
    <LineStyle>
      <color>ff0000ff</color><width>12</width>
    </LineStyle>
    <LabelStyle>
      <color>ff1f1fff</color><scale>3.5</scale>
    </LabelStyle>
    <IconStyle>
      <color>ff3f3fff</color><scale>2.0</scale><Icon><href>i.png</href></Icon>
    </IconStyle>
    <BalloonStyle>
      <text><![CDATA[
        <b>$[l]</b><br/>
        <table>
          <tr><td>Area</td><td>$[a]&nbsp;Mm²</td></tr>
          <tr><td>Ave. degree measure</td><td>$[k]&nbsp;Mm²</td></tr>
          <tr><td>Ave. clustering measure</td><td nowrap>$[C] %</td></tr>
          <tr><td>Ave. centrality measure</td><td nowrap>$[CC] %</td></tr>
          <tr><td>Ave. betweenness measure</td><td nowrap>$[B] %</td></tr>
          <tr><td>contained in region</td><td><a href="#p$[p];balloonFlyto">"""
                """$[pl]</a></td></tr>
          <tr><td>largest contained subregions</td><td><a href="#p$[n1];"""
                """balloonFlyto">$[l1]</a>, <a href="#p$[n2];balloonFlyto">"""
                """$[l2]</a></td></tr>
        </table>
      ]]></text>
    </BalloonStyle>
  </Style>
  <Style id="e"><!-- arcs -->
    <IconStyle>
      <color>00000000</color>
    </IconStyle>
    <BalloonStyle>
      <text><![CDATA[
        <b><a href="#p$[n1];balloonFlyto">$[l1]</a> - <a href="#p$[n2];"""
                """balloonFlyto">$[l2]</a></b><br/>
        <table>
          <tr><td>Strength (total product of linked area)</td><td nowrap>"""
                """$[w] (Mm²)²</td></tr>
          <tr><td>Relative strength</td><td nowrap>$[r] %</td></tr>
          <tr><td>Ave. network distance</td><td nowrap>$[l] steps</td></tr>
          <tr><td>Ave. link betweenness measure</td><td nowrap>$[B] %</td></tr>
        </table>
      ]]></text>
    </BalloonStyle>
  </Style>
  <Style id="f"><!-- highlit arcs -->
    <LineStyle>
      <color>ff0000ff</color><width>4</width>
    </LineStyle>
    <LabelStyle>
      <color>ff3f3fff</color><scale>2.5</scale>
    </LabelStyle>
    <IconStyle>
      <color>00000000</color>
    </IconStyle>
    <BalloonStyle>
      <text><![CDATA[
        <b><a href="#p$[n1];balloonFlyto">$[l1]</a> - <a href="#p$[n2];"""
                """balloonFlyto">$[l2]</a></b><br/>
        <table>
          <tr><td>Strength (total product of linked area)</td><td nowrap>"""
                """$[w] (Mm²)²</td></tr>
          <tr><td>Relative strength</td><td nowrap>$[r] %</td></tr>
          <tr><td>Ave. network distance</td><td nowrap>$[l] steps</td></tr>
          <tr><td>Ave. link betweenness measure</td><td nowrap>$[B] %</td></tr>
        </table>
      ]]></text>
    </BalloonStyle>
  </Style>
  <Folder id="regions">
    <visibility>0</visibility>""")
        intermediate = set(self.regions)
        intermediate_links = set(links)
        for leve, s in enumerate(self.solutions):
            f.write("""
    <Folder id="regions"""+str(s)+"""">""")
            sn = set(self.solution_nodes[level])
            for i in sn.intersection(intermediate):
                self._write_kml_region(f, i)
            intermediate -= sn
            f.write("""
    </Folder>""")
        f.write("""
    <Folder id="regions0">""")
        for i in intermediate:
            self._write_kml_region(f, i)
        f.write("""
    </Folder>
  </Folder>
  <Folder id="links">
    <visibility>0</visibility>""")
        limax = -np.inf
        for i, j in links:
            limax = max(limax, self.linked_weight[i, j])
        for level, s in enumerate(self.solutions):
            f.write("""
    <Folder id="links"""+str(s)+"""">""")
            for i in self.solution_nodes[level]:
                for j in self.solution_nodes[level]:
                    if (i, j) in intermediate_links:
                        intermediate_links.remove((i, j))
                        self._write_kml_link(f, i, j,
                                             self.linked_weight[i, j]/limax)
            f.write("""
    </Folder>""")
        f.write("""
    <Folder id="links0">""")
        for i, j in intermediate_links:
            self._write_kml_link(f, i, j, self.linked_weight[i, j]/limax)
        f.write("""
    </Folder>
  </Folder>
 </Document>
</kml>""")
        f.close()
        os.system("zip -9m "+path+"/data.kmz "+path+"/data.kml")

    def _write_kml_region(self, f, i):
        if self.parent[i] in self.regions:
            p = str(self.parent[i])
            pl = self.label[self.parent[i]]
        else:
            p = pl = ""
        if len(self.children[i]) > 1:
            c1 = self.children[i][0]
            if c1 in self.regions:
                l1 = self.label[c1]
                c1 = str(c1)
            else:
                l1 = c1 = ""
            c2 = self.children[i][1]
            if c2 in self.regions:
                l2 = self.label[c2]
                c2 = str(c2)
            else:
                l2 = c2 = ""
        else:
            l1 = c1 = l2 = c2 = ""
        f.write("""
<Placemark id="p"""+str(i)+""""><name/><ExtendedData><Data name="l"><value>
"""+self.label[i]+"""
</value></Data><Data name="a"><value>
"""+str(round(self.weight[i]*51007)/100)+"""
</value></Data><Data name="k"><value>
"""+str(round(self.k[i]*51007)/100)+"""
</value></Data><Data name="C"><value>
"""+str(int(100*self.C[i]))+"""
</value></Data><Data name="CC"><value>
"""+str(int(100*self.CC[i]))+"""
</value></Data><Data name="B"><value>
"""+str(round(10000*self.B[i])/100)+"""
</value></Data><Data name="p"><value>
"""+p+"""
</value></Data><Data name="pl"><value>
"""+pl+"""
</value></Data><Data name="n1"><value>
"""+c1+"""
</value></Data><Data name="l1"><value>
"""+l1+"""
</value></Data><Data name="n2"><value>
"""+c2+"""
</value></Data><Data name="l2"><value>
"""+l2+"""
</value></Data></ExtendedData><TimeSpan/><LookAt><gx:TimeStamp><when/>"""
                """</gx:TimeStamp><longitude>
"""+str(self.lon[i])+"""
</longitude><latitude>
"""+str(self.lat[i])+"""
</latitude><range>12000000</range></LookAt><styleUrl>#a</styleUrl><StyleMap>"""
                """<Pair><key>normal</key><Style><IconStyle id="i"""+str(i)
                + """"><color/>
<scale>1</scale><Icon><href>i.png</href></Icon></IconStyle><LabelStyle id="l"""
                + str(i)+""""><color>
ff"""+self.color[i][1:7]+"""
</color><scale>0</scale></LabelStyle><LineStyle><color>
ff"""+self.color[i][1:7]+"""
</color><width>6</width></LineStyle></Style></Pair><Pair><key>highlight"""
                """</key><styleUrl>#b</styleUrl></Pair></StyleMap>"""
                """<MultiGeometry><Point><coordinates>
"""+str(self.lon[i])+","+str(self.lat[i])+"""
</coordinates></Point>""")
        if self.shape[i] is not None:
            for sh in self.shape[i]:
                f.write("""
<Polygon><tessellate>1</tessellate><outerBoundaryIs>"""
                        """<LinearRing><coordinates>""")
                for la, lo in sh:
                    f.write("\n"+str(round(100*lo)/100) +
                            "," + str(round(100*la)/100))
                f.write("""
</coordinates></LinearRing></outerBoundaryIs></Polygon>""")
        f.write("""
</MultiGeometry></Placemark>""")

    def _write_kml_link(self, f, i, j, relwidth=0.1):
        nps = 21
        posi = np.array(
            [self.network.latlon2cartesian(self.lat[i], self.lon[i])])
        posj = np.array(
            [self.network.latlon2cartesian(self.lat[j], self.lon[j])])
        pos = posi + (posj-posi) * np.linspace(0, 1, nps).reshape((-1, 1))
        hgt = (.25 - (.5-np.linspace(0, 1, nps))**2) * 10000000 * \
            (0.05+((posi-posj)**2).sum()/4*0.95)
        pos /= np.sqrt((pos**2).sum(axis=1)).reshape((-1, 1))
        la, lo = self.network.cartesian2latlon(pos[nps/2, :])
        hotpos = str(round(100*lo)/100) + "," + str(round(100*la)/100) + \
            "," + str(round(hgt[nps/2]))
        d = self.D[i, j]
        if d > 0:
            d = str(round(d*100)/100)
        else:
            d = ""
        lb = self.LB[i, j]
        if lb > 0:
            lb = str(round(1000000*lb)/10000)
        else:
            lb = ""
        f.write("""
<Placemark id="p"""+str(i)+"-"+str(j)
                + """"><name/><ExtendedData><Data name="w"><value>
"""+str(round(self.linked_weight[i, j]*510.07**2*100)/100)+"""
</value></Data><Data name="r"><value>
"""+str(round(self.linked_proportion[i, j]*10000)/100)+"""
</value></Data><Data name="l"><value>
"""+d+"""
</value></Data><Data name="B"><value>
"""+lb+"""
</value></Data><Data name="n1"><value>
"""+str(i)+"""
</value></Data><Data name="l1"><value>
"""+self.label[i]+"""
</value></Data><Data name="n2"><value>
"""+str(i)+"""
</value></Data><Data name="l2"><value>
"""+self.label[j]+"""
</value></Data></ExtendedData><TimeSpan/><styleUrl>#e</styleUrl><StyleMap>"""
                """<Pair><key>normal</key><Style><LabelStyle id="l"""+str(i)
                + "-"+str(j)
                + """"><color>bf7fffff</color><scale>0</scale></LabelStyle>"""
                """<LineStyle id="a"""+str(i)+"-"+str(j)+""""><color/><width>
"""+str(relwidth*5+1)+"""
</width></LineStyle></Style></Pair><Pair><key>highlight</key><styleUrl>#f"""
                """</styleUrl></Pair></StyleMap><MultiGeometry><Point>"""
                """<altitudeMode>relativeToGround</altitudeMode><coordinates>
"""+hotpos+"""
</coordinates></Point><LineString><altitudeMode>relativeToGround"""
                """</altitudeMode><coordinates>""")
        for s in range(nps):
            la, lo = self.network.cartesian2latlon(pos[s, :])
            f.write("\n" + str(round(100*lo)/100) +
                    "," + str(round(100*la)/100) + "," + str(int(hgt[s])))
        f.write("""
</coordinates></LineString></MultiGeometry></Placemark>""")

    def _improve_step(self):
        """improve layout by one step"""
        # TODO: fruchterman or some such
        for i in self.shown:
            posi = self.position[i]
            for j, line in self.lines[i].items():
                if j > i:
                    posj = posi
                    # disterr = (((posi-posj)**2).sum()-1000000.0)**2
        self._update()

    def _process_expand_button_on(self, i):
        """expand a shown supernode and show collapsed descendants"""
        if i is None or i < self.N:
            return
        self._expand(i)
        self._hide(i)
        S = set([i])
        while len(S) > 0:
            for c in self.children[S.pop()]:
                if self.is_expanded[c]:
                    S.add(c)
                else:
                    # TODO: compute x,y more sensibly
                    if self.position[c, 0] == 0.0 and not self.fixed:
                        r = 4*self.radius[i] - self.radius[c]
                        phi = 2*np.pi * random.uniform()  # c/self.S
                        pos = self.position[i, :] + \
                            r * np.array([np.cos(phi), np.sin(phi)])
                        self._show(c, pos)
                    else:
                        self._show(c, self.position[c])

    def _process_collapse_button_on(self, i):
        """collapse parent of a shown supernode and hide descendants"""
        if i is None or i == self.S-1:
            return
        p = self.parent[i]
        self._collapse(p)
        S = set([p])
        sumweights = 0.0
        sumpos = np.zeros((self.dim))
        while len(S) > 0:
            for c in self.children[S.pop()]:
                if self.is_expanded[c]:
                    S.add(c)
                else:
                    sumweights += self.weight[c]
                    sumpos += self.weight[c] * self.position[c]
                    self._hide(c)
        if self.position[p, 0] == 0.0 and not self.fixed:
            self._show(p, sumpos/sumweights)
        else:
            self._show(p, self.position[p])

    def _show(self, i, pos):
        """
        show a supernode at some position and incident edges to other shown
        supernodes
        """
        if self.is_shown[i]:
            return False
        self.is_shown[i] = True
        self.shown.add(i)
        self.position[i] = pos
        color = self.color[i]
        self.circles[i] = mpl.pyplot.Circle(
            (pos[0], pos[1]), radius=self.radius[i], color=color, alpha=0.5)
        self.ax.add_patch(self.circles[i])
        if self.map:
            self.polys[i] = []
            for sh in self.shape[i]:
                sh = np.array(sh)
                x, y = self.bm(sh[:, 1], sh[:, 0])
                dists2 = (x[1:]-x[:-1])**2 + (y[1:]-y[:-1])**2
                jumps = np.where(dists2 > 50*dists2.min())[0]
                if len(jumps) > 1:
                    self.polys[i] += \
                        self.bm.plot(list(sh[1+jumps[-1]:, 1]) +
                                     list(sh[:1+jumps[0], 1]),
                                     list(sh[1+jumps[-1]:, 0]) +
                                     list(sh[:1+jumps[0], 0]),
                                     color=color, alpha=0.25, lw=3.0)
                    for j in range(len(jumps)-1):
                        self.polys[i] += \
                            self.bm.plot(sh[1+jumps[j]:1+jumps[j+1], 1],
                                         sh[1+jumps[j]:1+jumps[j+1], 0],
                                         color=color, alpha=0.25, lw=3.0)
                else:
                    self.polys[i] += \
                        self.bm.plot(sh[:, 1], sh[:, 0],
                                     color=color, alpha=0.25, lw=3.0)
        for j in self.shown:
            if j == i:
                continue
            lw = self.linked_weight[i, j]
            if lw > 0.0:
                color = "gray"
                self.lines[i][j] = self.lines[j][i] = line \
                    = mpl.pyplot.Line2D(self.position[[i, j], 0],
                                        self.position[[i, j], 1], alpha=0.5,
                                        color=color, linewidth=0.5+lw*99)
                self.ax.add_artist(line)
        self.fig.show()
        return True

    def _hide(self, i):
        """hide a supernode and incident edges"""
        if not self.is_shown[i]:
            return False
        self.is_shown[i] = False
        self.shown.remove(i)
        self.circles[i].remove()
        self.circles[i] = None
        if self.is_geo:
            for p in self.polys[i]:
                p.remove()
                del p
            self.polys[i] = []
        for j, line in self.lines[i].items():
            line.remove()
            del self.lines[j][i]
        self.lines[i] = {}
        self.fig.show()
        return True

    def _move(self, i, pos):
        """move a shown supernode to some position and adjust incident edges"""
        self.position[i] = pos
        self.circles[i].center = (pos[0], pos[1])
        for j, line in self.lines[i].items():
            line.set_data(self.position[[i, j], :].T)
        self.fig.show()

    def _update(self):
        for i in self.shown:
            pos = self.position[i]
            self.circles[i].center = (pos[0], pos[1])
            for j, line in self.lines[i].items():
                if j > i:
                    line.set_data(self.position[[i, j], :].T)
        self.fig.show()

    def _expand(self, i):
        """expand a supernode and return whether its state changed"""
        if self.is_expanded[i]:
            return False
        self.is_expanded[i] = True
        return True

    def _collapse(self, i):
        """collapse a supernode and return whether its state changed"""
        if not self.is_expanded[i]:
            return False
        self.is_expanded[i] = False
        return True


#     def _dummy():
#         f.write("""<?xml version="1.0" encoding="UTF-8"?>
# <kml xmlns="http://www.opengis.net/kml/2.2">
#  <Document>
#   <Style id="a"><!-- common to regions -->
#    <PolyStyle>
#     <fill>0</fill>
#     <outline>1</outline>
#    </PolyStyle>
#    <IconStyle>
#     <color>00000000</color>
#    </IconStyle>
#    <BalloonStyle>
#     <text><![CDATA[
#      <b>$[name]</b><br/>
#      <table>
#       <tr><td>Area</td><td><b>a</b></td><td nowrap>$[a] Mm²</td></tr>
#       <tr><td>Ave. linked area (n.s.i. degree)</td><td><b>k</b></td>"""
#                 """<td nowrap>$[k] Mm²</td></tr>
#       <tr><td>Ave. n.s.i. clustering coeff.</td><td><b>C</b></td>"""
#                 """<td nowrap>$[C] %</td></tr>
#       <tr><td>Ave. n.s.i. shortest path betweenness</td><td><b>B</b></td>"""
#                 """<td nowrap>$[B] %</td></tr>
#       <tr><td>contained in region</td><td/><td>$[p]</td></tr>
#       <tr><td>used in solution(s) no.</td><td/><td>$[t]</td></tr>
#       <tr><td>contains regions</td><td/><td>$[ch]</td></tr>
#      </table>
#     ]]></text>
#    </BalloonStyle>
#   </Style>
#   <Style id="b"><!-- highlit regions -->
#    <PolyStyle>
#     <fill>0</fill>
#     <outline>1</outline>
#    </PolyStyle>
#    <LineStyle>
#     <color>ff0000ff</color>
#     <width>12</width>
#    </LineStyle>
#    <LabelStyle>
#     <color>ff3f3fff</color>
#     <scale>4.0</scale>
#    </LabelStyle>
#    <IconStyle>
#     <color>00000000</color>
#    </IconStyle>
#    <BalloonStyle>
#     <text><![CDATA[
#      <b>$[name]</b><br/>
#      <table>
#       <tr><td>Area</td><td><b>a</b></td><td>$[a]&nbsp;Mm²</td></tr>
#       <tr><td>Ave. linked area (n.s.i. degree)</td><td><b>k</b></td><td>"""
#                 """$[k]&nbsp;Mm²</td></tr>
#       <tr><td>Ave. n.s.i. clustering coeff.</td><td><b>C</b></td>"""
#                 """<td nowrap>$[C] %</td></tr>
#       <tr><td>Ave. n.s.i. shortest path betweenness</td><td><b>B</b></td>"""
#                 """<td nowrap>$[B] %</td></tr>
#       <tr><td>contained in region</td><td/><td>$[p]</td></tr>
#       <tr><td>used in solution(s) no.</td><td/><td>$[t]</td></tr>
#       <tr><td>contains regions</td><td/><td>$[ch]</td></tr>
#      </table>
#     ]]></text>
#    </BalloonStyle>
#   </Style>
#   <Style id="c"><!-- common to nodes -->
#    <IconStyle>
#     <color>00000000</color>
#    </IconStyle>
#   </Style>
#   <Style id="d"><!-- highlit nodes -->
#    <LabelStyle>
#     <color>ff3f3fff</color>
#     <scale>4.0</scale>
#    </LabelStyle>
#    <IconStyle>
#     <color>00000000</color>
#    </IconStyle>
#   </Style>
#   <Style id="e"><!-- arcs -->
#    <LabelStyle>
#     <color>bf7fffff</color>
#     <scale>1.0</scale>
#    </LabelStyle>
#    <IconStyle>
#     <color>00000000</color>
#    </IconStyle>
#    <BalloonStyle/>
#     <!--<text><![CDATA[
#       <b>$[name]</b><br/>
#       <table>
#       <tr><td>Strength</td><td><b>a</b></td><td nowrap>$[s] (Mm²)²</td></tr>
#       <tr><td>Relative strength</td><td><b>a</b></td><td nowrap>$[r] %"""
#                 """</td></tr>
#       <tr><td>Ave. network distance</td><td><b>l</b></td><td nowrap>"""
#                 """$[l] steps</td></tr>
#       <tr><td>Ave. n.s.i. shortest path betweenness</td><td><b>B</b></td>"""
#                 """<td nowrap>$[B] %</td></tr>
#       <tr><td>used in solution(s) no.</td><td/><td>$[t]</td></tr>
#       </table>
#     ]]></text>
#    </BalloonStyle>-->
#   </Style>
#   <Style id="f"><!-- highlit arcs -->
#    <LineStyle>
#     <color>ff0000ff</color>
#     <width>4</width>
#    </LineStyle>
#    <LabelStyle>
#     <color>ff3f3fff</color>
#     <scale>1.5</scale>
#    </LabelStyle>
#    <IconStyle>
#     <color>00000000</color>
#     </IconStyle>
#    <BalloonStyle/>
#     <!--<text><![CDATA[
#       <b>$[name]</b><br/>
#       <table>
#       <tr><td>Strength</td><td><b>a</b></td><td nowrap>$[s] (Mm²)²</td></tr>
#       <tr><td>Relative strength</td><td><b>a</b></td><td nowrap>$[r] %"""
#                 """</td></tr>
#       <tr><td>Ave. network distance</td><td><b>l</b></td><td nowrap>"""
#                 """$[l] steps</td></tr>
#       <tr><td>Ave. n.s.i. shortest path betweenness</td><td><b>B</b></td>"""
#                 """<td nowrap>$[B] %</td></tr>
#       <tr><td>used in solution(s) no.</td><td/><td>$[t]</td></tr>
#       </table>
#     ]]></text>
#    </BalloonStyle>-->
#   </Style>
#   <Style id="h"><!-- hidden children -->
#    <ListStyle>
#     <listItemType>checkHideChildren</listItemType>
#    </ListStyle>
#   </Style>
#   <Style id="l">
#    <ListStyle>
#     <ItemIcon/>
#    </ListStyle>
#   </Style>
#   <Style id="lh">
#    <ListStyle>
#     <listItemType>checkHideChildren</listItemType>
#    </ListStyle>
#   </Style>
#   <Style id="lr">
#    <ListStyle>
#     <listItemType>radioFolder</listItemType>
#    </ListStyle>
#   </Style>
#   <Folder id="0">
#    <name>"""+title+"""</name>
#    <styleUrl>#lr</styleUrl>
#    <Folder id="h">
#     <name>Slideable Hierarchy of Regions</name>
#     <description><![CDATA[
#      region labels inside subfolders give this information:<br/>
#      <b>round
#       <a href="#h;balloon">code</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"""
#                 """&nbsp;&nbsp;&nbsp;&nbsp;
#       <a href="#h;balloon">a</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"""
#                 """&nbsp;&nbsp;&nbsp;
#       <a href="#h;balloon">k</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"""
#                 """&nbsp;&nbsp;&nbsp;
#       <a href="#h;balloon">C</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"""
#                 """&nbsp;&nbsp;&nbsp;
#       <a href="#h;balloon">B</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"""
#                 """&nbsp;&nbsp;&nbsp;
#       <a href="#h;balloon">location</a>
#      </b><br/><br/>
#      <table>
#       <tr>
#        <td nowrap><i>
#         code<br/>
#         a<br/>
#         k<br/>
#         C<br/>
#         B<br/>
#         location
#        </i></td><td nowrap>
#         is a short unique code,<br/>
#         is the region's area,<br/>
#         is the region's average linked area (n.s.i. degree),<br/>
#         is the region's average n.s.i. clustering coeff.,<br/>
#         is the region's average n.s.i. shortest path betweenness,<br/>
#         gives its approximate location.
#        </td>
#       </tr>
#      </table>
#      <h1>Slideable Hierarchy of Regions</h1>
#      <p>This folder contains a hierarchy of regions covering the Earth's
#      surface. Use the slider on the top-left to successively increase the
#      level of detail from a coarse partition into """
#      +str(len(self.solution_nodes[0]))+""" regions to a
#      fine partition into """+str(len(self.solution_nodes[-1]))+""" regions
#      (if the slider shows a "month" and "day", just ignore them).
#      <p>Expand this folder to get more information on individual regions,
#      ordered into subfolders corresponding to certain selected levels of
#      detail. The hierarchy can be used in combination with third-party
#      overlays that do not themselves use the slider. (For third-party
#      overlays using the slider, use the folder """
#      """<a href="#p;balloonFlyto">Selected Partitions</a> instead.)
#     ]]></description>
#     <visibility>1</visibility>""")
#         f.write("""
#     </Folder>
#    </Folder>
#    <Folder id="p">
#     <name>Selected Partitions</name>
#     <description><![CDATA[
#      region labels inside subfolders give this information:<br/>
#      <b>
#       <a href="#p;balloon">code</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"""
#                 """&nbsp;&nbsp;&nbsp;&nbsp;
#       <a href="#p;balloon">a</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"""
#                 """&nbsp;&nbsp;&nbsp;
#       <a href="#p;balloon">k</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"""
#                 """&nbsp;&nbsp;&nbsp;
#       <a href="#p;balloon">C</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"""
#                 """&nbsp;&nbsp;&nbsp;
#       <a href="#p;balloon">B</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"""
#                 """&nbsp;&nbsp;&nbsp;
#       <a href="#p;balloon">location</a>
#      </b><br/><br/>
#      <table>
#       <tr>
#        <td nowrap><i>
#         code<br/>
#         a<br/>
#         k<br/>
#         C<br/>
#         B<br/>
#         location
#        </i></td><td nowrap>
#         is a short unique code,<br/>
#         is the region's area,<br/>
#         is the region's average linked area (n.s.i. degree),<br/>
#         is the region's average n.s.i. clustering coeff.,<br/>
#         is the region's average n.s.i. shortest path betweenness,<br/>
#         gives its approximate location.
#        </td>
#       </tr>
#      </table>
#      <h1>Selected Partitions</h1>
#      <p>This folder contains sets of regions that each cover the Earth's
#      surface and correspond to certain selected levels of detail from the
#      above hierarchy. They can be used in combination with third-party
#      overlays that use a slider. Expand the subfolders to get more
#      information on individual regions.
#      <a href="r5063;balloonFlyto">TEST</a>
#     ]]></description>
#     <styleUrl>#lr</styleUrl>""")



# Copyright (C) 2014 SWIPO Project
#
# Authors (this file):
#   Stefan Schinkel <stefan.schinkel@gmail.com>
"""
Provides sanity checks for basic for parallel and serial circiuts.
"""
import numpy as np
import networkx as nx


from pyunicorn import ResNetwork
from .ResistiveNetwork_utils import \
    makeNW, parallelCopy, serialCopy, nx2nw

debug = 0
""" Test for basic sanity, parallel and serial circiuts
"""


def testParallelTrivial():
    r""" Trivial parallel case:
    a)  0 --- 1 --- 2

        /---- 3 ---\
    b)  0 --- 1 --- 2

    c)  /---- 3 ---\
        0 --- 1 --- 2
        \____ 4 ___/

    ER(a) = 2*ER(b) = 3*ER(c)
    """
    nws = []
    # construct nw1
    idI, idJ = [0, 1], [1, 2]
    nws.append(makeNW(idI, idJ, [.1]))

    # construct nw2
    idI += [0, 3]
    idJ += [3, 2]
    nws.append(makeNW(idI, idJ, [.1]))

    # nw3
    idI += [0, 4]
    idJ += [4, 2]
    nws.append(makeNW(idI, idJ, [.1]))

    ER = []
    for nw in nws:
        rnw = ResNetwork(nw)
        ER.append(rnw.effective_resistance(0, 2))

    assert abs(ER[0]/2-ER[1]) < .1E-6
    assert abs(ER[0]/3-ER[2]) < .1E-6


def testParallelLessTrivial():
    """ Less Trivial Parallel Case:
            |--- 1 --- 0
        a) 2     |
            |--- 3 ----4

            |--- 1 --- 0 --- 5 --- |
        b) 2     |           |     7
            |--- 3 ----4 --- 6 --- |

                      |---- 8 ----------- |
                      |     |             |
                      |     |----------|  |
                      |                |  |
           |--- 1 --- 0 --- 5 --- |    |  |
        c) 2    |           |     7    |  9
           |--- 3 ----4 --- 6 --- |    |  |
                      |                |  |
                      |      ----------|  |
                      |      |            |
                      |---- 10 -----------|
    """
    nws = []

    idI = [0, 1, 1, 2, 3]
    idJ = [1, 2, 3, 3, 4]
    nws.append(makeNW(idI, idJ, [1]*len(idI)))

    idI.extend([0, 5, 5, 6, 6])
    idJ.extend([5, 6, 7, 7, 4])
    nws.append(makeNW(idI, idJ, [1]*len(idI)))

    idI.extend([0, 8, 8, 9, 10])
    idJ.extend([8, 9, 10, 10, 4])
    nws.append(makeNW(idI, idJ, [1]*len(idI)))

    ER = []
    for nw in nws:
        rnw = ResNetwork(nw)
        ER.append(rnw.effective_resistance(0, 4))
    #     Gs.append(nx.DiGraph(nw))
    # # showGraphs(Gs)
    # # s = ''
    # # for i,e in enumerate(ER):
    # #     s = s + "NW{:d} {:.3f}\t".format(i,e)
    # # print("Effective resistances (0,2)\n %s" % (s))

    assert abs(ER[0]/2-ER[1]) < .1E-6
    assert abs(ER[0]/3-ER[2]) < .1E-6

    # """ Less Trivial Parallel Case:
    #     /--- 1 --- 0
    # a) 2     |
    #     \--- 3 ----4

    #     /--- 1 --- 0 --- 5 --- \
    # b) 2     |           |      7
    #     \--- 3 ----4 --- 6 --- /

    #               / --- 8 ----------- \
    #               |                    \
    #    /--- 1 --- 0 --- 5 --- \         \
    # c) 2                       7        9
    #    \--- 3 ----4 --- 6 --- /         /
    #               |                    /
    #               \ --- 10 -----------/
    # """
    # nws =[]
    # #construct nw1

    # idI = [0,1,1,2,3]
    # idJ = [1,2,3,3,4]
    # val = [.1] * 5
    # nws.append(makeNW(idI,idJ,[.1]*len(idI))[0])

    # idI.extend([0,5,6,7])
    # idJ.extend([5,6,7,4])
    # val.extend( val * 6)
    # nws.append(makeNW(idI,idJ,[.1]*len(idI))[0])

    # idI.extend([0,8,9,10])
    # idJ.extend([8,9,10,4])
    # val.extend( val * 4)
    # nws.append(makeNW(idI,idJ,val)[0])

    # ER = []
    # for nw in nws:
    #     rnw = ResNetwork(nw)
    #     ER.append( rnw.effective_resistance(0,4))

    # s = ''
    # for i,e in enumerate(ER):
    #     s = s + "NW{:d} {:.3f}\t".format(i,e)
    # print("Effective resistances (0,2)\n %s" % (s))

    # assert abs(ER[0]/2-ER[1]) < .1E-6
    # assert abs(ER[0]/3-ER[2]) < .1E-6


def testParallelRandom():
    """ 50 random parallel cases
    """

    N = 10
    p = .7

    runs = 0
    while runs < 50:

        G = nx.fast_gnp_random_graph(N, p)
        a = 0
        b = G.number_of_nodes()-1

        try:
            nx.shortest_path(G, source=a, target=b)
        except RuntimeError:
            continue

        i, j = [], []
        for xx in G.edges():
            i.append(xx[0])
            j.append(xx[1])

        # %.1f values for resistance
        val = np.round(np.random.ranf(len(i))*100)/10

        # and test
        nw1 = makeNW(i, j, val)
        nw2 = parallelCopy(nw1, a, b)
        ER1 = ResNetwork(nw1).effective_resistance(a, b)
        ER2 = ResNetwork(nw2).effective_resistance(a, b)

        # assertion
        assert (ER1/2-ER2) < 1E-6

        # increment runs
        runs += 1


def testSerialTrivial():
    """Trivial serial test case

    a) 0 --- 1 --- 2

    b) 0 --- 1 --- 2 --- 3 --- 4

    ER(a)/2 = ER(b)
    """

    # construct nw1
    idI = [0, 1]
    idJ = [1, 2]
    val = [1, 1]

    nw1 = np.zeros((3, 3))
    for i, j, v in zip(idI, idJ, val):
        nw1[i, j] = v
        nw1[j, i] = v

    # construct nw2
    idI = idI + [2, 3]
    idJ = idJ + [3, 4]
    val = val + [1, 1]

    nw2 = np.zeros((5, 5))
    for i, j, v in zip(idI, idJ, val):
        nw2[i, j] = v
        nw2[j, i] = v

    # init ResNetworks
    rnw1 = ResNetwork(nw1)
    rnw2 = ResNetwork(nw2)

    ER1 = rnw1.effective_resistance(0, 2)
    ER2 = rnw2.effective_resistance(0, 4)

    print("Effective resistances (0,2)")
    print(f"NW1 {ER1:.3f}\tNW2 {ER2:.3f}\t 2*NW1 = {(2*ER1):.3f}")

    assert (ER1*2-ER2) < 1E-6


def testSerialRandom():
    """ 50 Random serial test cases
    """

    N = 10
    p = .7
    runs = 50
    for run in range(0, runs):

        # a random graph
        G = nx.fast_gnp_random_graph(N, p)
        try:
            nx.shortest_path(G, source=0, target=N-1)
        except RuntimeError:
            continue
        except nx.NetworkXNoPath:
            pass
        # convert to plain ndarray
        nw1 = nx2nw(G)

        # copy and join network
        nw2 = serialCopy(nw1)

        # compute effective resistance
        ER1 = ResNetwork(
            nw1, silence_level=3).effective_resistance(0, len(nw1)-1)
        ER2 = ResNetwork(
            nw2, silence_level=3).effective_resistance(0, len(nw2)-1)

        # assertion
        # print(ER1*2-ER2)
        assert (ER1*2-ER2) < 1E-6
